import decimal
import itertools


def get_properties(schema):
    '''
    Given a schema dict, return its properties. But if it's got sub-schemas with multiple different
    potential properties, return their merged properties instead (interleaved so the first
    properties of each sub-schema come first). The idea is that the user should see all possible
    options even if they're not all possible together.
    '''
    if 'oneOf' in schema:
        return dict(
            item
            for item in itertools.chain(
                *itertools.zip_longest(
                    *[sub_schema['properties'].items() for sub_schema in schema['oneOf']],
                ),
            )
            if item is not None
        )

    return schema.get('properties', {})


SCHEMA_TYPE_TO_PYTHON_TYPE = {
    'array': list,
    'boolean': bool,
    'integer': int,
    'number': decimal.Decimal,
    'object': dict,
    'string': str,
}


def parse_type(schema_type, **overrides):
    '''
    Given a schema type as a string, return the corresponding Python type.

    If any overrides are given in the from of a schema type string to a Python type, then override
    the default type mapping with them.

    Raise ValueError if the schema type is unknown.
    '''
    try:
        return dict(
            SCHEMA_TYPE_TO_PYTHON_TYPE,
            **overrides,
        )[schema_type]
    except KeyError:
        raise ValueError(f'Unknown type in configuration schema: {schema_type}')


def compare_types(schema_type, target_types, match=any):
    '''
    Given a schema type as a string or a list of strings (representing multiple types) and a set of
    target type strings, return whether every schema type is in the set of target types.

    If the schema type is a list of strings, use the given match function (such as any or all) to
    compare elements. For instance, if match is given as all, then every element of the schema_type
    list must be in the target types.
    '''
    if isinstance(schema_type, list):
        return match(element_schema_type in target_types for element_schema_type in schema_type)

    return schema_type in target_types



import collections
import locale
import logging
import os
import pathlib
import shutil
import tempfile

import borgmatic.actions.dump
import borgmatic.actions.pattern
import borgmatic.borg.extract
import borgmatic.borg.list
import borgmatic.borg.repo_list
import borgmatic.config.paths
import borgmatic.hooks.data_source.dump
import borgmatic.hooks.dispatch

logger = logging.getLogger(__name__)


UNSPECIFIED = object()


Dump = collections.namedtuple(
    'Dump',
    ('hook_name', 'data_source_name', 'hostname', 'port', 'label', 'container'),
    defaults=(None, None, None, None),
)


def dumps_match(first, second, default_port=None):
    '''
    Compare two Dump instances for equality while supporting a field value of UNSPECIFIED, which
    indicates that the field should match any value. If a default port is given, then consider any
    dump having that port to match with a dump having a None port.
    '''
    # label kinda counts as an unique id, if they match ignore host/container/port
    if first.label not in {None, UNSPECIFIED} and first.label == second.label:
        field_list = ('hook_name', 'data_source_name')
    else:
        field_list = Dump._fields

    for field_name in field_list:
        first_value = getattr(first, field_name)
        second_value = getattr(second, field_name)

        if default_port is not None and field_name == 'port':
            if first_value == default_port and second_value is None:
                continue

            if second_value == default_port and first_value is None:
                continue

        if first_value == UNSPECIFIED or second_value == UNSPECIFIED:  # noqa: PLR1714
            continue

        if first_value != second_value:
            return False

    return True


def render_dump_metadata(dump):
    '''
    Given a Dump instance, make a display string describing it for use in log messages.
    '''
    label = dump.label or UNSPECIFIED
    name = 'unspecified' if dump.data_source_name is UNSPECIFIED else dump.data_source_name
    host = dump.container or dump.hostname or UNSPECIFIED
    port = None if dump.port is UNSPECIFIED else dump.port

    if label is not UNSPECIFIED:
        metadata = f'{name}@{label}'
    elif port:
        metadata = f'{name}@:{port}' if host is UNSPECIFIED else f'{name}@{host}:{port}'
    else:
        metadata = f'{name}' if host is UNSPECIFIED else f'{name}@{host}'

    if dump.hook_name not in {None, UNSPECIFIED}:
        return f'{metadata} ({dump.hook_name})'

    return metadata


def get_configured_data_source(config, restore_dump):
    '''
    Search in the given configuration dict for dumps corresponding to the given dump to restore. If
    there are multiple matches, error.

    Return the found data source as a data source configuration dict or None if not found.
    '''
    try:
        hooks_to_search = {restore_dump.hook_name: config[restore_dump.hook_name]}
    except KeyError:
        return None

    matching_dumps = tuple(
        hook_data_source
        for (hook_name, hook_config) in hooks_to_search.items()
        for hook_data_source in hook_config
        for default_port in (
            borgmatic.hooks.dispatch.call_hook(
                function_name='get_default_port',
                config=config,
                hook_name=hook_name,
            ),
        )
        if dumps_match(
            Dump(
                hook_name,
                hook_data_source.get('name'),
                hook_data_source.get('hostname'),
                hook_data_source.get('port'),
                hook_data_source.get('label') or UNSPECIFIED,
                hook_data_source.get('container'),
            ),
            restore_dump,
            default_port,
        )
    )

    if not matching_dumps:
        return None

    if len(matching_dumps) > 1:
        raise ValueError(
            f'Cannot restore data source {render_dump_metadata(restore_dump)} because there are multiple matching data sources configured',
        )

    return matching_dumps[0]


def strip_path_prefix_from_extracted_dump_destination(
    extract_path,
    borgmatic_runtime_directory,
):
    '''
    Directory-format dump files get extracted into a temporary directory containing a path prefix
    that depends how the files were stored in the archive. So, given the path where the dump was
    extracted and the borgmatic runtime directory, move the dump files such that the restore doesn't
    have to deal with that varying path prefix.

    For instance, if the dump was extracted to:

      /run/user/0/borgmatic/tmp1234/borgmatic/postgresql_databases/test/...

    or:

      /run/user/0/borgmatic/tmp1234/root/.borgmatic/postgresql_databases/test/...

    then this function moves it to:

      /run/user/0/borgmatic/postgresql_databases/test/...
    '''
    for subdirectory_path, _, _ in os.walk(extract_path):
        databases_directory = os.path.basename(subdirectory_path)

        if not databases_directory.endswith('_databases'):
            continue

        destination_path = os.path.join(borgmatic_runtime_directory, databases_directory)
        shutil.rmtree(destination_path, ignore_errors=True)
        shutil.move(subdirectory_path, destination_path)

        break


def restore_single_dump(
    repository,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    archive_name,
    hook_name,
    data_source,
    connection_params,
    borgmatic_runtime_directory,
):
    '''
    Given (among other things) an archive name, a data source hook name, the hostname, port,
    username/password as connection params, and a configured data source configuration dict, restore
    that data source from the archive.
    '''
    dump_metadata = render_dump_metadata(
        Dump(
            hook_name,
            data_source['name'],
            data_source.get('hostname'),
            data_source.get('port'),
            data_source.get('label') or UNSPECIFIED,
            data_source.get('container'),
        ),
    )

    logger.info(f'Restoring data source {dump_metadata}')

    dump_patterns = borgmatic.hooks.dispatch.call_hooks(
        'make_data_source_dump_patterns',
        config,
        borgmatic.hooks.dispatch.Hook_type.DATA_SOURCE,
        borgmatic_runtime_directory,
        data_source['name'],
        data_source.get('hostname'),
        data_source.get('port'),
        data_source.get('container'),
        data_source.get('label'),
    )[hook_name.split('_databases', 1)[0]]

    destination_path = (
        tempfile.mkdtemp(dir=borgmatic_runtime_directory)
        if data_source.get('format') == 'directory'
        else None
    )

    try:
        # Kick off a single data source extract. If using a directory format, extract to a temporary
        # directory. Otherwise extract the single dump file to stdout.
        extract_process = borgmatic.borg.extract.extract_archive(
            dry_run=global_arguments.dry_run,
            repository=repository['path'],
            archive=archive_name,
            paths=[
                borgmatic.hooks.data_source.dump.convert_glob_patterns_to_borg_pattern(
                    dump_patterns,
                ),
            ],
            config=config,
            local_borg_version=local_borg_version,
            global_arguments=global_arguments,
            local_path=local_path,
            remote_path=remote_path,
            destination_path=destination_path,
            # A directory format dump isn't a single file, and therefore can't extract
            # to stdout. In this case, the extract_process return value is None.
            extract_to_stdout=bool(data_source.get('format') != 'directory'),
        )

        if destination_path and not global_arguments.dry_run:
            strip_path_prefix_from_extracted_dump_destination(
                destination_path,
                borgmatic_runtime_directory,
            )
    finally:
        if destination_path and not global_arguments.dry_run:
            shutil.rmtree(destination_path, ignore_errors=True)

    # Run a single data source restore, consuming the extract stdout (if any).
    borgmatic.hooks.dispatch.call_hook(
        function_name='restore_data_source_dump',
        config=config,
        hook_name=hook_name,
        data_source=data_source,
        dry_run=global_arguments.dry_run,
        extract_process=extract_process,
        connection_params=connection_params,
        borgmatic_runtime_directory=borgmatic_runtime_directory,
    )


def collect_dumps_from_archive(
    repository,
    archive,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    borgmatic_runtime_directory,
):
    '''
    Given a local or remote repository path, a resolved archive name, a configuration dict, the
    local Borg version, global arguments an argparse.Namespace, local and remote Borg paths, and the
    borgmatic runtime directory, query the archive for the names of data sources dumps it contains
    and return them as a tuple of Dump instances.
    '''
    dumps_from_archive = {}  # Use a dict as an ordered set.

    # There is (at most) one dump metadata file per data source hook. Load each.
    for dumps_metadata_entry in borgmatic.borg.list.capture_archive_listing(
        repository,
        archive,
        config,
        local_borg_version,
        global_arguments,
        list_paths=[
            'sh:'
            + borgmatic.hooks.data_source.dump.make_data_source_dump_path(
                base_directory,
                '*_databases/dumps.json',
            )
            # Probe for dump metadata files in multiple locations, as the default location is
            # "/borgmatic/*_databases/dumps.json" with Borg 1.4+, but instead begins with the
            # borgmatic runtime directory for older versions of Borg.
            for base_directory in (
                'borgmatic',
                borgmatic.config.paths.make_runtime_directory_glob(borgmatic_runtime_directory),
            )
        ],
        local_path=local_path,
        remote_path=remote_path,
    ):
        if not dumps_metadata_entry.get('path'):
            continue

        for dump in borgmatic.hooks.data_source.dump.parse_data_source_dumps_metadata(
            borgmatic.borg.extract.extract_archive(
                global_arguments.dry_run,
                repository,
                archive,
                [dumps_metadata_entry['path']],
                config,
                local_borg_version,
                global_arguments,
                local_path=local_path,
                remote_path=remote_path,
                extract_to_stdout=True,
            )
            .stdout.read()
            .decode(locale.getpreferredencoding()),
            dumps_metadata_entry['path'],
        ):
            dumps_from_archive[dump] = None

    # If we've successfully loaded any dumps metadata, we're done.
    if dumps_from_archive:
        logger.debug('Collecting database dumps from archive data source dumps metadata files')

        return tuple(dumps_from_archive.keys())

    # No dumps metadata files were found, so for backwards compatibility, fall back to parsing the
    # paths of dumps found in the archive to get their respective dump metadata.
    logger.debug('Collecting database dumps from archive data source dump paths (fallback)')
    borgmatic_source_directory = str(
        pathlib.Path(borgmatic.config.paths.get_borgmatic_source_directory(config)),
    )

    # Probe for the data source dumps in multiple locations, as the default location has moved to
    # the borgmatic runtime directory (which gets stored as just "/borgmatic" with Borg 1.4+). But
    # we still want to support reading dumps from previously created archives as well.
    dump_entries = borgmatic.borg.list.capture_archive_listing(
        repository,
        archive,
        config,
        local_borg_version,
        global_arguments,
        list_paths=[
            'sh:'
            + borgmatic.hooks.data_source.dump.make_data_source_dump_path(
                base_directory,
                '*_databases/*/*',
            )
            for base_directory in (
                'borgmatic',
                borgmatic.config.paths.make_runtime_directory_glob(borgmatic_runtime_directory),
                borgmatic_source_directory.lstrip('/'),
            )
        ],
        local_path=local_path,
        remote_path=remote_path,
    )

    for dump_entry in dump_entries:
        dump_path = dump_entry.get('path')

        if not dump_path:
            continue

        # Probe to find the base directory that's at the start of the dump path.
        for base_directory in (
            'borgmatic',
            borgmatic_runtime_directory,
            borgmatic_source_directory,
        ):
            try:
                (hook_name, host_and_port, data_source_name) = dump_path.split(
                    base_directory + os.path.sep,
                    1,
                )[1].split(os.path.sep)[0:3]
            except (ValueError, IndexError):
                continue

            parts = host_and_port.split(':', 1)

            if len(parts) == 1:
                parts += (None,)

            (hostname, port) = parts

            try:
                port = int(port)
            except (ValueError, TypeError):
                port = None

            dumps_from_archive[
                Dump(
                    hook_name, data_source_name, None if hostname == 'localhost' else hostname, port
                )
            ] = None

            # We've successfully parsed the dump path, so need to probe any further.
            break
        else:
            logger.warning(
                f'Ignoring invalid data source dump path "{dump_path}" in archive {archive}',
            )

    return tuple(dumps_from_archive.keys())


def get_dumps_to_restore(restore_arguments, dumps_from_archive):
    '''
    Given restore arguments as an argparse.Namespace instance indicating which dumps to restore and
    a tuple of Dump instances representing the dumps found in an archive, return a tuple of specific
    Dump instances from the archive to restore. As part of this, replace any Dump having a data
    source name of "all" with multiple named Dump instances as appropriate.

    Raise ValueError if any of the requested data source names cannot be found in the archive or if
    there are multiple archive dump matches for a given requested dump.
    '''
    requested_dumps = tuple(
        # Use a dict comprehension as an ordered set.
        {
            Dump(
                hook_name=(
                    (
                        restore_arguments.hook
                        if restore_arguments.hook.endswith('_databases')
                        else f'{restore_arguments.hook}_databases'
                    )
                    if restore_arguments.hook
                    else UNSPECIFIED
                ),
                data_source_name=name,
                hostname=restore_arguments.original_hostname or UNSPECIFIED,
                port=restore_arguments.original_port,
                label=restore_arguments.original_label or UNSPECIFIED,
                container=restore_arguments.original_container or UNSPECIFIED,
            ): None
            for name in restore_arguments.data_sources or (UNSPECIFIED,)
        }.keys()
        if restore_arguments.hook
        or restore_arguments.data_sources
        or restore_arguments.original_hostname
        or restore_arguments.original_port
        or restore_arguments.original_label
        or restore_arguments.original_container
        else (
            Dump(
                hook_name=UNSPECIFIED,
                data_source_name='all',
                hostname=UNSPECIFIED,
                port=UNSPECIFIED,
                label=UNSPECIFIED,
                container=UNSPECIFIED,
            ),
        )
    )
    missing_dumps = set()
    dumps_to_restore = {}  # Use a dict as an ordered set.

    # If there's a requested "all" dump, add every dump from the archive to the dumps to restore.
    if any(dump for dump in requested_dumps if dump.data_source_name == 'all'):
        for dump in dumps_from_archive:
            dumps_to_restore[dump] = None

    # If any archive dump matches a requested dump, add the archive dump to the dumps to restore.
    for requested_dump in requested_dumps:
        if requested_dump.data_source_name == 'all':
            continue

        matching_dumps = tuple(
            archive_dump
            for archive_dump in dumps_from_archive
            if dumps_match(requested_dump, archive_dump)
        )

        if len(matching_dumps) == 0:
            missing_dumps.add(requested_dump)
        elif len(matching_dumps) == 1:
            dumps_to_restore[matching_dumps[0]] = None
        else:
            raise ValueError(
                f'Cannot restore data source {render_dump_metadata(requested_dump)} because there are multiple matching dumps in the archive. Try adding flags to disambiguate.',
            )

    if missing_dumps:
        rendered_dumps = ', '.join(
            f'{render_dump_metadata(dump)}' for dump in sorted(missing_dumps)
        )

        raise ValueError(
            f"Cannot restore data source dump{'s' if len(missing_dumps) > 1 else ''} {rendered_dumps} missing from archive",
        )

    return tuple(dumps_to_restore.keys())


def ensure_requested_dumps_restored(dumps_to_restore, dumps_actually_restored):
    '''
    Given a tuple of requested dumps to restore and a set of dumps actually restored, raise
    ValueError if any requested dumps to restore weren't restored, indicating that they were missing
    from the configuration.
    '''
    if not dumps_to_restore:
        raise ValueError('No data source dumps were found to restore')

    missing_dumps = sorted(
        set(dumps_to_restore) - dumps_actually_restored,
        key=lambda dump: dump.data_source_name,
    )

    if missing_dumps:
        rendered_dumps = ', '.join(f'{render_dump_metadata(dump)}' for dump in missing_dumps)

        raise ValueError(
            f"Cannot restore data source{'s' if len(missing_dumps) > 1 else ''} {rendered_dumps} missing from borgmatic's configuration",
        )


def run_restore(
    repository,
    config,
    local_borg_version,
    restore_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "restore" action for the given repository.

    Raise ValueError if a configured data source could not be found to restore or there's no
    matching dump in the archive.
    '''
    logger.info(f'Restoring data sources from archive {restore_arguments.archive}')
    working_directory = borgmatic.config.paths.get_working_directory(config)

    with borgmatic.config.paths.Runtime_directory(config) as borgmatic_runtime_directory:
        patterns = borgmatic.actions.pattern.process_patterns(
            borgmatic.actions.pattern.collect_patterns(config, working_directory),
            config,
            working_directory,
        )

        with borgmatic.actions.dump.Dump_cleanup(
            config, borgmatic_runtime_directory, patterns, global_arguments.dry_run
        ):
            archive_name = borgmatic.borg.repo_list.resolve_archive_name(
                repository['path'],
                restore_arguments.archive,
                config,
                local_borg_version,
                global_arguments,
                local_path,
                remote_path,
            )
            dumps_from_archive = collect_dumps_from_archive(
                repository['path'],
                archive_name,
                config,
                local_borg_version,
                global_arguments,
                local_path,
                remote_path,
                borgmatic_runtime_directory,
            )
            dumps_to_restore = get_dumps_to_restore(restore_arguments, dumps_from_archive)

            dumps_actually_restored = set()
            connection_params = {
                'container': restore_arguments.container,
                'hostname': restore_arguments.hostname,
                'port': restore_arguments.port,
                'username': restore_arguments.username,
                'password': restore_arguments.password,
                'restore_path': restore_arguments.restore_path,
            }

            # Restore each dump.
            for restore_dump in dumps_to_restore:
                found_data_source = get_configured_data_source(
                    config,
                    restore_dump,
                )
                # For a dump that wasn't found via an exact match in the configuration, try to fallback
                # to an "all" data source.
                if not found_data_source:
                    found_data_source = get_configured_data_source(
                        config,
                        Dump(
                            restore_dump.hook_name,
                            'all',
                            restore_dump.hostname,
                            restore_dump.port,
                            restore_dump.label,
                            restore_dump.container,
                        ),
                    )

                    if not found_data_source:
                        continue

                    found_data_source = dict(found_data_source)
                    found_data_source['name'] = restore_dump.data_source_name

                dumps_actually_restored.add(restore_dump)

                restore_single_dump(
                    repository,
                    config,
                    local_borg_version,
                    global_arguments,
                    local_path,
                    remote_path,
                    archive_name,
                    restore_dump.hook_name,
                    dict(found_data_source, schemas=restore_arguments.schemas),
                    connection_params,
                    borgmatic_runtime_directory,
                )

    ensure_requested_dumps_restored(dumps_to_restore, dumps_actually_restored)



import logging
import os
import signal
import sys

logger = logging.getLogger(__name__)


EXIT_CODE_FROM_SIGNAL = 128


def handle_signal(signal_number, frame):
    '''
    Send the signal to all processes in borgmatic's process group, which includes child processes.
    '''
    # Prevent infinite signal handler recursion. If the parent frame is this very same handler
    # function, we know we're recursing.
    if frame.f_back.f_code.co_name == handle_signal.__name__:
        return

    os.killpg(os.getpgrp(), signal_number)

    if signal_number == signal.SIGTERM:
        logger.critical('Exiting due to TERM signal')
        sys.exit(EXIT_CODE_FROM_SIGNAL + signal.SIGTERM)
    elif signal_number == signal.SIGINT:
        # Borg doesn't always exit on a SIGINT, so give it a little encouragement.
        os.killpg(os.getpgrp(), signal.SIGTERM)

        raise KeyboardInterrupt()


def configure_signals():
    '''
    Configure borgmatic's signal handlers to pass relevant signals through to any child processes
    like Borg.
    '''
    for signal_number in (
        signal.SIGHUP,
        signal.SIGINT,
        signal.SIGTERM,
        signal.SIGUSR1,
        signal.SIGUSR2,
    ):
        signal.signal(signal_number, handle_signal)



import logging
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, flags
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


def transfer_archives(
    dry_run,
    repository_path,
    config,
    local_borg_version,
    transfer_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a dry-run flag, a local or remote repository path, a configuration dict, the local Borg
    version, the arguments to the transfer action, and global arguments as an argparse.Namespace
    instance, transfer archives to the given repository.
    '''
    borgmatic.logger.add_custom_log_levels()
    extra_borg_options = config.get('extra_borg_options', {}).get('transfer', '')

    full_command = (
        (local_path, 'transfer')
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + flags.make_flags('remote-path', remote_path)
        + flags.make_flags('umask', config.get('umask'))
        + (('--log-json',) if (config.get('log_json') or not config.get('progress')) else ())
        + flags.make_flags('lock-wait', config.get('lock_wait'))
        + flags.make_flags('progress', config.get('progress'))
        + (
            flags.make_flags_from_arguments(
                transfer_arguments,
                excludes=(
                    'repository',
                    'source_repository',
                    'archive',
                    'match_archives',
                    'progress',
                ),
            )
            or (
                flags.make_match_archives_flags(
                    transfer_arguments.archive or config.get('match_archives'),
                    config.get('archive_name_format'),
                    local_borg_version,
                )
            )
        )
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
        + flags.make_flags('other-repo', transfer_arguments.source_repository)
        + flags.make_flags('dry-run', dry_run)
    )

    return execute_command(
        full_command,
        output_log_level=logging.ANSWER,
        output_file=DO_NOT_CAPTURE if config.get('progress') else None,
        environment=environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import logging
import os

import textual
import textual.signal

import borgmatic.actions.browse.archive

logger = logging.getLogger('__name__')


LOADING_DONE = object()


class Archive_loaded(textual.signal.Signal):
    '''
    A signal that publishes when each subsequent archive is loaded from a repository, intended for
    consumption in widgets that display archives as they are loaded. This signal also publishes
    when loading is complete.

    Each subscribed callback call includes the archive as an archive name string. Given the lack of
    other identifying information (configuration file, repository), there should be a separate
    Archive_loaded instance per repository.
    '''


@textual.work(thread=True)
def add_repository_archives(browse_app, archive_loaded, config, repository):
    '''
    Given a running Browse_app instance, an Archive_loaded instance, a configuration dict, and a
    repository dict, load a list of the archives from the repository and add them as options in the
    archives list. Reverse the order so the most recent archive is first.

    This function runs in a separate thread from the main UI. When loading is complete, publish a
    loading done signal.
    '''
    archives_data = borgmatic.actions.browse.archive.get_repository_archives(config, repository)

    # Reverse the archives, so the common case of accessing the latest archive is easy because it's
    # at the top.
    for archive in reversed(archives_data['archives']):
        archive_loaded.publish(archive['archive'])

    archive_loaded.publish(LOADING_DONE)


def record_path(archive_path, hierarchy, path_components):
    '''
    Given an Archive_path instance, a dict capturing a filesystem hierarchy of paths, and a tuple of
    path components for the archive path, set the archive path into the hierarchy data structure.

    For instance, if given an archive path and path components representing "foo/bar/baz.txt",
    produce a hierarchy that looks like:

        {'foo': {'bar': {'baz.txt': Archive_path('-', 'foo/bar/baz.txt', '')}}}

    Note that the hierarchy is modified in place, so any existing paths there are retained.
    '''
    if len(path_components) == 1:
        hierarchy[path_components[0]] = {} if archive_path.path_type == 'd' else archive_path
        return

    record_path(archive_path, hierarchy.setdefault(path_components[0], {}), path_components[1:])


def get_paths(hierarchy, path_components, full_path_components=None):
    '''
    Given a dict capturing a filesystem hierarchy of paths (or a subset thereof), a tuple of path
    components for a directory path relative to the hierarchy root, and an optional tuple of
    *absolute* path components for the same path (if different), return a generator of the
    contained file and directory Archive_path instances from the hierarchy.

    For instance, given the following hierarchy:

        {'foo': {'bar': {'baz.txt': Archive_path('-', 'foo/bar/baz.txt', ''), 'quux': {}}}}

    ... and path components of ('foo', 'bar'), return a generator with the following:

        * Archive_path('-', 'foo/bar/baz.txt', '')
        * Archive_path('d', 'foo/bar/quux', '')

    The given absolute path components are use to construct directory paths like that last archive
    path.
    '''
    if full_path_components is None:
        full_path_components = path_components

    if len(path_components) == 1:
        try:
            return (
                archive_path
                if isinstance(archive_path, borgmatic.actions.browse.archive.Archive_path)
                else borgmatic.actions.browse.archive.Archive_path(
                    'd', os.path.join(*full_path_components, component), ''
                )
                for component, archive_path in hierarchy[path_components[0]].items()
            )
        except KeyError:
            raise ValueError(f'Unknown file or directory: {path_components[0]}')

    try:
        return get_paths(hierarchy[path_components[0]], path_components[1:], full_path_components)
    except KeyError:
        raise ValueError(f'Unknown directory: {path_components[0]}')


class Archive_path_loaded(textual.signal.Signal):
    '''
    A signal that publishes when each subsequent path is loaded from an archive, intended for
    consumption in widgets that display paths as they are loaded. This signal also tracks the
    complete filesystem hierarchy seen thus far, so new widgets that get created after loading has
    started can "catch up" with existing known paths. Lastly, this signal publishes and tracks when
    loading is complete.

    Each subscrided callback call includes the loaded path as an Archive_path instance. There is
    intended to be a separate Archive_path_loaded instance per archive, but that instance should be
    shared among several different widgets for the same archive for performance reasons.
    '''

    def __init__(self, owner, name):
        self.path_hierarchy = {}
        self.complete = False

        super().__init__(owner, name)

    def publish(self, archive_path):
        '''
        Publish the given archive path to subscribers and record its path locally. But if the
        archive path is actually LOADING_DONE, then record loading as complete.
        '''
        super().publish(archive_path)

        if archive_path is LOADING_DONE:
            self.complete = True
        else:
            record_path(
                archive_path, self.path_hierarchy, archive_path.file_path.split(os.path.sep)
            )


@textual.work(thread=True)
def load_archive_paths(browse_app, path_loaded, config, repository, archive_name):
    '''
    Given a running Browse_app instance, an Archive_path_loaded instance, a configuration dict, a
    repository dict, and an archive name, load the paths in this archive and publish each one via
    the Archive_path_loaded signal, so interested widgets can subscribe. Also send a "loading done"
    signal when loading completes.

    This function runs in a separate thread from the main UI. When loading is complete, publish a
    loading done signal.
    '''
    for archive_path in borgmatic.actions.browse.archive.get_archive_paths(
        config, repository, archive_name
    ):
        path_loaded.publish(archive_path)

    path_loaded.publish(LOADING_DONE)


class File_preview_loaded(textual.signal.Signal):
    '''
    A signal that publishes when file contents are loaded from an archive, intended for consumption
    in widgets that display loaded files. This signal also publishes when loading is complete.

    Each published callback includes a the file's contents as a string. Given the lack of other
    identifying information (configuration file, repository, archive), there should be a separate
    Archive_loaded instance per previewed file.
    '''


@textual.work(thread=True)
def load_file_preview(
    browse_app,
    file_preview_loaded,
    config,
    repository,
    archive_name,
    file_path,
):
    '''
    Given a running Browse_app instance, a File_preview_loaded instance, a configuration dict, a
    repository dict, an archive name, and the path of a file in that archive, load the contents of
    the file and write it into the given file preview widget.

    This function runs in a separate thread from the main UI. When loading is complete, publish a
    loading done signal.
    '''
    file_contents = borgmatic.actions.browse.archive.get_archive_file_content(
        config, repository, archive_name, file_path
    )

    file_preview_loaded.publish(file_contents)



import logging
import re

import requests

import borgmatic.hooks.monitoring.logs
from borgmatic.hooks.monitoring import monitor

logger = logging.getLogger(__name__)

MONITOR_STATE_TO_HEALTHCHECKS = {
    monitor.State.START: 'start',
    monitor.State.FINISH: None,  # Healthchecks doesn't append to the URL for the finished state.
    monitor.State.FAIL: 'fail',
    monitor.State.LOG: 'log',
}

DEFAULT_PING_BODY_LIMIT_BYTES = 100000
HANDLER_IDENTIFIER = 'healthchecks'
TIMEOUT_SECONDS = 10


def initialize_monitor(hook_config, config, config_filename, monitoring_log_level, dry_run):
    '''
    Add a handler to the root logger that stores in memory the most recent logs emitted. That way,
    we can send them all to Healthchecks upon a finish or failure state. But skip this if the
    "send_logs" option is false.
    '''
    if hook_config.get('send_logs', False) is False:
        return

    ping_body_limit = max(
        hook_config.get('ping_body_limit', DEFAULT_PING_BODY_LIMIT_BYTES)
        - len(borgmatic.hooks.monitoring.logs.PAYLOAD_TRUNCATION_INDICATOR),
        0,
    )

    borgmatic.hooks.monitoring.logs.add_handler(
        borgmatic.hooks.monitoring.logs.Forgetful_buffering_handler(
            HANDLER_IDENTIFIER,
            ping_body_limit,
            monitoring_log_level,
        ),
    )


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Ping the configured Healthchecks URL or UUID, modified with the monitor.State. Use the given
    configuration filename in any log entries, and log to Healthchecks with the giving log level.
    If this is a dry run, then don't actually ping anything.
    '''
    ping_url = (
        hook_config['ping_url']
        if hook_config['ping_url'].startswith('http')
        else f"https://hc-ping.com/{hook_config['ping_url']}"
    )
    dry_run_label = ' (dry run; not actually pinging)' if dry_run else ''

    if 'states' in hook_config and state.name.lower() not in hook_config['states']:
        logger.info(f'Skipping Healthchecks {state.name.lower()} ping due to configured states')
        return

    ping_url_is_uuid = re.search(r'\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$', ping_url)

    healthchecks_state = MONITOR_STATE_TO_HEALTHCHECKS.get(state)
    if healthchecks_state:
        ping_url = f'{ping_url}/{healthchecks_state}'

    if hook_config.get('create_slug'):
        if ping_url_is_uuid:
            logger.warning('Healthchecks UUIDs do not support auto provisionning; ignoring')
        else:
            ping_url = f'{ping_url}?create=1'

    logger.info(f'Pinging Healthchecks {state.name.lower()}{dry_run_label}')
    logger.debug(f'Using Healthchecks ping URL {ping_url}')

    if state in {monitor.State.FINISH, monitor.State.FAIL, monitor.State.LOG}:
        payload = borgmatic.hooks.monitoring.logs.format_buffered_logs_for_payload(
            HANDLER_IDENTIFIER,
        )
    else:
        payload = ''

    if not dry_run:
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        try:
            response = requests.post(
                ping_url,
                data=payload.encode('utf-8'),
                verify=hook_config.get('verify_tls', True),
                timeout=TIMEOUT_SECONDS,
                headers={'User-Agent': 'borgmatic'},
            )
            if not response.ok:
                response.raise_for_status()
        except requests.exceptions.RequestException as error:
            logger.warning(f'Healthchecks error: {error}')


def destroy_monitor(hook_config, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    Remove the monitor handler that was added to the root logger. This prevents the handler from
    getting reused by other instances of this monitor.
    '''
    borgmatic.hooks.monitoring.logs.remove_handler(HANDLER_IDENTIFIER)



import logging

import borgmatic.borg.borg
import borgmatic.borg.repo_list

logger = logging.getLogger(__name__)


def run_borg(
    repository,
    config,
    local_borg_version,
    borg_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "borg" action for the given repository.
    '''
    logger.info('Running arbitrary Borg command')
    archive_name = borgmatic.borg.repo_list.resolve_archive_name(
        repository['path'],
        borg_arguments.archive,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
    )
    borgmatic.borg.borg.run_arbitrary_borg(
        repository['path'],
        config,
        local_borg_version,
        options=borg_arguments.options,
        archive=archive_name,
        local_path=local_path,
        remote_path=remote_path,
    )



import logging

import borgmatic.actions.json
import borgmatic.borg.repo_list

logger = logging.getLogger(__name__)


def run_repo_list(
    repository,
    config,
    local_borg_version,
    repo_list_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "repo-list" action for the given repository.

    If repo_list_arguments.json is True, yield the JSON output from listing the repository.
    '''
    if not repo_list_arguments.json:
        logger.answer('Listing repository')

    json_output = borgmatic.borg.repo_list.list_repository(
        repository['path'],
        config,
        local_borg_version,
        repo_list_arguments=repo_list_arguments,
        global_arguments=global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )
    if json_output:
        yield borgmatic.actions.json.parse_json(json_output, repository.get('label'))



import logging
import os
import pathlib
import shlex
import stat
import textwrap

import borgmatic.borg.pattern
import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, feature, flags
from borgmatic.execute import (
    DO_NOT_CAPTURE,
    execute_command,
    execute_command_and_capture_output,
    execute_command_with_processes,
)

logger = logging.getLogger(__name__)


def special_file(path, working_directory=None):
    '''
    Return whether the given path is a special file (character device, block device, or named pipe
    / FIFO). If a working directory is given, take it into account when making the full path to
    check.
    '''
    try:
        mode = os.stat(os.path.join(working_directory or '', path)).st_mode
    except (FileNotFoundError, OSError):
        return False

    return stat.S_ISCHR(mode) or stat.S_ISBLK(mode) or stat.S_ISFIFO(mode)


def any_parent_directories(path, candidate_parents):
    '''
    Return whether any of the given candidate parent directories are an actual parent of the given
    path. This includes grandparents, etc.
    '''
    for parent in candidate_parents:
        if pathlib.PurePosixPath(parent) in pathlib.PurePath(path).parents:
            return True

    return False


def validate_planned_backup_paths(
    dry_run,
    create_command,
    config,
    patterns,
    local_path,
    working_directory,
    borgmatic_runtime_directory,
    find_special_files=False,
):
    '''
    Given a dry-run flag, a Borg create command as a tuple, a configuration dict, a local Borg path,
    a working directory, and the borgmatic runtime directory, perform a "borg create --dry-run" to
    determine whether Borg's planned paths to include in a backup look good. Specifically, if the
    given runtime directory exists, validate that it will be included in a backup and hasn't been
    excluded.

    If find special files is True, then return the subset of planned backup paths that are special
    files. Otherwise, return an empty tuple.

    Raise ValueError if the runtime directory has been excluded via "exclude_patterns" or similar,
    because any features that rely on the runtime directory getting backed up will break. For
    instance, without the runtime directory, Borg can't consume any database dumps and borgmatic may
    hang waiting for them to be consumed.
    '''
    # Omit "--exclude-nodump" from the Borg dry run command, because that flag causes Borg to open
    # files including any named pipe we've created. And omit "--filter" because that can break the
    # paths output parsing below such that path lines no longer start with the expected "- ".
    path_lines = execute_command_and_capture_output(
        (
            *flags.omit_flag_and_value(
                flags.omit_flag(
                    flags.omit_flag(create_command, '--exclude-nodump'),
                    '--log-json',
                ),
                '--filter',
            ),
            '--dry-run',
            '--list',
        ),
        capture_stderr=True,
        working_directory=working_directory,
        environment=environment.make_environment(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )

    # These are all the individual files that Borg is planning to backup as determined by the Borg
    # create dry run above.
    paths = (
        path_line.split(' ', 1)[1]
        for path_line in path_lines
        if path_line and path_line.startswith(('- ', '+ '))
    )

    include_pattern_paths = {
        pattern.path
        for pattern in patterns
        if pattern.type == borgmatic.borg.pattern.Pattern_type.INCLUDE
    }
    runtime_directory_root_patterns = tuple(
        pattern
        for pattern in patterns
        if any_parent_directories(pattern.path, (borgmatic_runtime_directory,))
        if pattern.type == borgmatic.borg.pattern.Pattern_type.ROOT
        # Skip root patterns that have corresponding include patterns, because those will "punch
        # through" any subsequent excludes.
        if pattern.path not in include_pattern_paths
    )

    special_paths = []
    validate_runtime_directory = bool(
        not dry_run
        and os.path.exists(borgmatic_runtime_directory)
        and runtime_directory_root_patterns
    )
    runtime_directory_in_path = False

    # Do everything in this one loop because we only want to consume the paths generator once.
    for path in paths:
        # If all root patterns in the runtime directory are missing from the paths Borg is planning
        # to backup, then they must've gotten excluded, e.g. by user-configured excludes. Warn
        # accordingly (below).
        if validate_runtime_directory and any(
            any_parent_directories(path, (pattern.path,))
            for pattern in runtime_directory_root_patterns
        ):
            runtime_directory_in_path = True

        # Return the subset of output paths that are special files but *not* contained within the
        # borgmatic runtime directory. The intent is to skip runtime paths that borgmatic uses for its
        # own bookkeeping, instead focusing on user-configured paths.
        if (
            find_special_files
            and not any_parent_directories(path, (borgmatic_runtime_directory,))
            and special_file(path, working_directory)
        ):
            special_paths.append(path)

    if validate_runtime_directory and not runtime_directory_in_path:
        logger.warning(
            f'The runtime directory {os.path.normpath(borgmatic_runtime_directory)} overlaps with the configured excludes (or the snapshotted source directories are empty). Please ensure the runtime directory is not excluded.'
        )

    return tuple(special_paths)


MAX_SPECIAL_FILE_PATHS_LENGTH = 1000


def make_base_create_command(  # noqa: PLR0912
    dry_run,
    repository_path,
    config,
    patterns,
    local_borg_version,
    global_arguments,
    borgmatic_runtime_directory,
    archive_suffix='',
    local_path='borg',
    remote_path=None,
    json=False,
    comment=None,
    stream_processes=None,
):
    '''
    Given verbosity/dry-run flags, a local or remote repository path, a configuration dict, a
    sequence of patterns as borgmatic.borg.pattern.Pattern instances, the local Borg version, global
    arguments as an argparse.Namespace instance, the borgmatic runtime directory, a string suffix to
    add to the archive name, the local Borg path, the remote Borg path, whether to output JSON,
    comment text to add to the created archive, and a sequence of processes streaming data to Borg,
    return a tuple of (base Borg create command flags, Borg create command positional arguments,
    open pattern file handle).
    '''
    if config.get('source_directories_must_exist', True):
        borgmatic.borg.pattern.check_all_root_patterns_exist(
            patterns, borgmatic.config.paths.get_working_directory(config)
        )

    patterns_file = borgmatic.borg.pattern.write_patterns_file(
        patterns,
        borgmatic_runtime_directory,
    )
    checkpoint_interval = config.get('checkpoint_interval')
    checkpoint_volume = config.get('checkpoint_volume')
    chunker_params = config.get('chunker_params')
    compression = config.get('compression')
    upload_rate_limit = config.get('upload_rate_limit')
    upload_buffer_size = config.get('upload_buffer_size')
    umask = config.get('umask')
    lock_wait = config.get('lock_wait')
    list_filter_flags = flags.make_list_filter_flags(local_borg_version, dry_run)
    files_changed = config.get('files_changed')
    files_cache = config.get('files_cache')
    archive_name_format = (
        config.get('archive_name_format', flags.get_default_archive_name_format(local_borg_version))
        + archive_suffix
    )
    extra_borg_options = config.get('extra_borg_options', {}).get('create', '')

    if feature.available(feature.Feature.ATIME, local_borg_version):
        atime_flags = ('--atime',) if config.get('atime') is True else ()
    else:
        atime_flags = ('--noatime',) if config.get('atime') is False else ()

    if feature.available(feature.Feature.NOFLAGS, local_borg_version):
        noflags_flags = ('--noflags',) if config.get('flags') is False else ()
    else:
        noflags_flags = ('--nobsdflags',) if config.get('flags') is False else ()

    if feature.available(feature.Feature.NUMERIC_IDS, local_borg_version):
        numeric_ids_flags = ('--numeric-ids',) if config.get('numeric_ids') else ()
    else:
        numeric_ids_flags = ('--numeric-owner',) if config.get('numeric_ids') else ()

    if feature.available(feature.Feature.UPLOAD_RATELIMIT, local_borg_version):
        upload_ratelimit_flags = (
            ('--upload-ratelimit', str(upload_rate_limit)) if upload_rate_limit else ()
        )
    else:
        upload_ratelimit_flags = (
            ('--remote-ratelimit', str(upload_rate_limit)) if upload_rate_limit else ()
        )

    create_flags = (
        tuple(local_path.split(' '))
        + ('create',)
        + (('--patterns-from', patterns_file.name) if patterns_file else ())
        + flags.make_exclude_flags(config)
        + (('--comment', comment) if comment else ())
        + (('--checkpoint-interval', str(checkpoint_interval)) if checkpoint_interval else ())
        + (('--checkpoint-volume', str(checkpoint_volume)) if checkpoint_volume else ())
        + (('--chunker-params', chunker_params) if chunker_params else ())
        + (('--compression', compression) if compression else ())
        + upload_ratelimit_flags
        + (('--upload-buffer', str(upload_buffer_size)) if upload_buffer_size else ())
        + (('--one-file-system',) if config.get('one_file_system') else ())
        + numeric_ids_flags
        + atime_flags
        + (('--noctime',) if config.get('ctime') is False else ())
        + (('--nobirthtime',) if config.get('birthtime') is False else ())
        + (('--read-special',) if config.get('read_special') or stream_processes else ())
        + noflags_flags
        + (('--files-changed', files_changed) if files_changed else ())
        + (('--files-cache', files_cache) if files_cache else ())
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--log-json',) if (config.get('log_json') or not config.get('progress')) else ())
        + (
            ('--list', '--filter', list_filter_flags)
            if config.get('list_details') and not json and not config.get('progress')
            else ()
        )
        + (('--dry-run',) if dry_run else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
    )

    create_positional_arguments = flags.make_repository_archive_flags(
        repository_path,
        archive_name_format,
        local_borg_version,
    )
    working_directory = borgmatic.config.paths.get_working_directory(config)

    if config.get('unsafe_skip_path_validation_before_create'):
        logger.debug(
            'Skipping pre-backup path validation due to "unsafe_skip_path_validation_before_create" option.'
        )

        return (create_flags, create_positional_arguments, patterns_file)

    logger.debug('Checking file paths Borg plans to include')

    special_file_paths = validate_planned_backup_paths(
        dry_run,
        create_flags + create_positional_arguments,
        config,
        patterns,
        local_path,
        working_directory,
        borgmatic_runtime_directory=borgmatic_runtime_directory,
        find_special_files=bool(stream_processes),
    )

    if stream_processes and config.get('read_special') is False:
        logger.warning(
            'Ignoring configured "read_special" value of false, as true is needed for database hooks.',
        )

    # If database hooks are enabled (as indicated by streaming processes), exclude files that might
    # cause Borg to hang. But skip this if the user has explicitly set the "read_special" to True.
    if special_file_paths:
        truncated_special_file_paths = textwrap.shorten(
            ', '.join(special_file_paths),
            width=MAX_SPECIAL_FILE_PATHS_LENGTH,
            placeholder=' ...',
        )
        logger.warning(
            f'Excluding special files to prevent Borg from hanging: {truncated_special_file_paths}',
        )
        patterns_file = borgmatic.borg.pattern.write_patterns_file(
            tuple(
                borgmatic.borg.pattern.Pattern(
                    special_file_path,
                    borgmatic.borg.pattern.Pattern_type.NO_RECURSE,
                    borgmatic.borg.pattern.Pattern_style.FNMATCH,
                    source=borgmatic.borg.pattern.Pattern_source.INTERNAL,
                )
                for special_file_path in special_file_paths
            ),
            borgmatic_runtime_directory,
            patterns_file=patterns_file,
        )

        if '--patterns-from' not in create_flags:
            create_flags += ('--patterns-from', patterns_file.name)

    return (create_flags, create_positional_arguments, patterns_file)


def create_archive(
    dry_run,
    repository_path,
    config,
    patterns,
    local_borg_version,
    global_arguments,
    borgmatic_runtime_directory,
    archive_suffix='',
    local_path='borg',
    remote_path=None,
    json=False,
    comment=None,
    stream_processes=None,
):
    '''
    Given verbosity/dry-run flags, a local or remote repository path, a configuration dict, a
    sequence of loaded configuration paths, the local Borg version, global arguments as an
    argparse.Namespace instance, the borgmatic runtime directory, a string suffix to add to the
    archive name, the local Borg path, the remote Borg path, whether to output JSON, and comment
    text to add to the created archive, and a sequence of processes streaming data to Borg, create a
    Borg archive and return Borg's JSON output (if any).

    If a sequence of stream processes is given (instances of subprocess.Popen), then execute the
    create command while also triggering the given processes to produce output.
    '''
    borgmatic.logger.add_custom_log_levels()

    working_directory = borgmatic.config.paths.get_working_directory(config)

    (create_flags, create_positional_arguments, _) = make_base_create_command(
        dry_run,
        repository_path,
        config,
        patterns,
        local_borg_version,
        global_arguments,
        borgmatic_runtime_directory,
        archive_suffix,
        local_path,
        remote_path,
        json,
        comment,
        stream_processes,
    )

    if json:
        output_log_level = None
    elif config.get('list_details') or (
        (config.get('statistics') or config.get('quick_statistics')) and not dry_run
    ):
        output_log_level = logging.ANSWER
    else:
        output_log_level = logging.INFO

    # The progress output isn't compatible with captured and logged output, as progress messes with
    # the terminal directly.
    output_file = DO_NOT_CAPTURE if config.get('progress') else None

    create_flags += (
        (('--info',) if logger.getEffectiveLevel() == logging.INFO and not json else ())
        + (('--stats',) if config.get('statistics') and not json and not dry_run else ())
        + (
            ('--quick-stats',)
            if config.get('quick_statistics') and not json and not dry_run
            else ()
        )
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) and not json else ())
        + (('--progress',) if config.get('progress') else ())
        + (('--json',) if json else ())
    )
    borg_exit_codes = config.get('borg_exit_codes')

    if stream_processes:
        output = '\n'.join(
            execute_command_with_processes(
                create_flags + create_positional_arguments,
                stream_processes,
                output_log_level,
                output_file,
                working_directory=working_directory,
                environment=environment.make_environment(config),
                borg_local_path=local_path,
                borg_exit_codes=borg_exit_codes,
            )
        )
        return output if json else None

    if output_log_level is None:
        output = '\n'.join(
            execute_command_and_capture_output(
                create_flags + create_positional_arguments,
                working_directory=working_directory,
                environment=environment.make_environment(config),
                borg_local_path=local_path,
                borg_exit_codes=borg_exit_codes,
            )
        )
        return output if json else None

    execute_command(
        create_flags + create_positional_arguments,
        output_log_level,
        output_file,
        working_directory=working_directory,
        environment=environment.make_environment(config),
        borg_local_path=local_path,
        borg_exit_codes=borg_exit_codes,
    )

    return None



import signal

import textual.app
import textual.binding
import textual.widgets

import borgmatic.actions.browse.carousel
import borgmatic.actions.browse.configuration_files_list
import borgmatic.actions.browse.logs
import borgmatic.actions.browse.repositories_list


class Browse_app(textual.app.App):
    '''
    The main app / entry point for the browse action UI.
    '''

    BINDINGS = (
        textual.binding.Binding(key='q', action='quit', description='quit'),
        textual.binding.Binding(key='v', action='toggle_logs', description='view logs'),
        textual.binding.Binding(
            key='c', action='command_palette', description='commands', show=False
        ),
    )
    COMMAND_PALETTE_BINDING = 'c'
    CSS = '''
        .panel {
            border: round $primary;
            border-title-color: $text-primary;
            width: 100%;
            height: 100%;
        }

        #logs {
            width: 100%;
            height: 50%;
            display: none;
        }
    '''

    def __init__(self, configs):
        self.configs = configs

        super().__init__()

    def get_system_commands(self, screen):  # pragma: no cover
        '''
        Remove the screenshot system command because it produces broken screenshots. (Emoji
        weirdness, etc.) Also remove minimize and maximize because they don't accomplish much with
        out particular layout.
        '''
        yield from (
            command
            for command in super().get_system_commands(screen)
            if command.title not in {'Screenshot', 'Minimize', 'Maximize'}
        )

    def compose(self):
        '''
        Compose a UI consisting of:

          * a header with the application name
          * a carousel container that contains the main UI panels
          * a logs panel where Python logs show up (panel hidden by default)
          * a footer with available keys listed
        '''
        yield textual.widgets.Header()
        yield borgmatic.actions.browse.carousel.Carousel(
            [
                borgmatic.actions.browse.configuration_files_list.Configuration_files_list(
                    self.configs
                )
            ]
            if len(self.configs) > 1
            else [
                borgmatic.actions.browse.repositories_list.Repositories_list(
                    next(iter(self.configs.values()))
                )
            ]
        )

        logs_panel = borgmatic.actions.browse.logs.Logs()
        yield logs_panel
        yield textual.widgets.Footer()

        borgmatic.actions.browse.logs.log_to_widget(logs_panel)

    def on_mount(self):
        '''
        Set the application title, which ends up in the header.
        '''
        self.title = 'borgmatic browse'

    def action_toggle_logs(self):
        '''
        Toggle the show/hide status of the logs panel.
        '''
        logs_panel = self.query_one('#logs')
        logs_panel.styles.display = 'none' if logs_panel.styles.display == 'block' else 'block'

    def exit(self):  # pragma: no cover
        '''
        Exit the application. But first raise a SIGTERM (handled in borgmatic/signals.py) to
        encourage a fast exit by killing any ongoing Borg subprocesses.
        '''
        signal.raise_signal(signal.SIGTERM)

        super().exit()



import csv
import itertools
import logging
import os
import pathlib
import shlex

import borgmatic.borg.pattern
import borgmatic.config.paths
import borgmatic.hooks.credential.parse
import borgmatic.hooks.data_source.config
from borgmatic.execute import (
    execute_command,
    execute_command_and_capture_output,
    execute_command_with_processes,
)
from borgmatic.hooks.data_source import config as database_config
from borgmatic.hooks.data_source import dump

logger = logging.getLogger(__name__)


def make_dump_path(base_directory):  # pragma: no cover
    '''
    Given a base directory, make the corresponding dump path.
    '''
    return dump.make_data_source_dump_path(base_directory, 'postgresql_databases')


def make_environment(database, config, restore_connection_params=None):
    '''
    Make an environment dict from the current environment variables and the given database
    configuration. If restore connection params are given, this is for a restore operation.
    '''
    environment = dict(os.environ)

    password = database_config.resolve_database_option(
        'password', database, restore_connection_params, restore=restore_connection_params
    )

    if password:
        environment['PGPASSWORD'] = borgmatic.hooks.credential.parse.resolve_credential(
            password,
            config,
        )

    if 'ssl_mode' in database:
        environment['PGSSLMODE'] = database['ssl_mode']

    if 'ssl_cert' in database:
        environment['PGSSLCERT'] = database['ssl_cert']

    if 'ssl_key' in database:
        environment['PGSSLKEY'] = database['ssl_key']

    if 'ssl_root_cert' in database:
        environment['PGSSLROOTCERT'] = database['ssl_root_cert']

    if 'ssl_crl' in database:
        environment['PGSSLCRL'] = database['ssl_crl']

    return environment


EXCLUDED_DATABASE_NAMES = ('template0', 'template1')


def database_names_to_dump(database, config, environment, dry_run):
    '''
    Given a requested database config and a configuration dict, return the corresponding sequence of
    database names to dump. In the case of "all" when a database format is given, query for the
    names of databases on the configured host and return them. For "all" without a database format,
    just return a sequence containing "all".
    '''
    requested_name = database['name']

    if requested_name != 'all':
        return (requested_name,)

    if not database.get('format'):
        return ('all',)

    if dry_run:
        return ()

    psql_command = tuple(
        shlex.quote(part) for part in shlex.split(database.get('psql_command') or 'psql')
    )
    hostname = database_config.resolve_database_option('hostname', database)
    list_command = (
        psql_command
        + ('--list', '--no-password', '--no-psqlrc', '--csv', '--tuples-only')
        + (('--host', hostname) if hostname else ())
        + (('--port', str(database['port'])) if 'port' in database else ())
        + (
            (
                '--username',
                borgmatic.hooks.credential.parse.resolve_credential(database['username'], config),
            )
            if 'username' in database
            else ()
        )
        + (
            tuple(shlex.quote(part) for part in shlex.split(database['list_options']))
            if 'list_options' in database
            else ()
        )
    )
    logger.debug('Querying for "all" PostgreSQL databases to dump')
    list_lines = execute_command_and_capture_output(
        list_command,
        environment=environment,
        working_directory=borgmatic.config.paths.get_working_directory(config),
    )

    return tuple(
        row[0]
        for row in csv.reader(list_lines, delimiter=',', quotechar='"')
        if row[0] not in EXCLUDED_DATABASE_NAMES
    )


def get_default_port(databases, config):  # pragma: no cover
    return 5432


def use_streaming(databases, config):
    '''
    Given a sequence of PostgreSQL database configuration dicts, a configuration dict (ignored),
    return whether streaming will be using during dumps.
    '''
    return any(database.get('format') != 'directory' for database in databases)


def dump_data_sources(
    databases,
    config,
    config_paths,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):
    '''
    Dump the given PostgreSQL databases to a named pipe. The databases are supplied as a sequence of
    dicts, one dict describing each database as per the configuration schema. Use the given
    borgmatic runtime directory to construct the destination path.

    Return a sequence of subprocess.Popen instances for the dump processes ready to spew to a named
    pipe. But if this is a dry run, then don't actually dump anything and return an empty sequence.
    Also append the the parent directory of the database dumps to the given patterns list, so the
    dumps actually get backed up.

    Raise ValueError if the databases to dump cannot be determined.
    '''
    dry_run_label = ' (dry run; not actually dumping anything)' if dry_run else ''
    processes = []
    dumps_metadata = []

    logger.info(f'Dumping PostgreSQL databases{dry_run_label}')
    dump_path = make_dump_path(borgmatic_runtime_directory)

    for database in databases:
        environment = make_environment(database, config)
        dump_database_names = database_names_to_dump(database, config, environment, dry_run)

        if not dump_database_names:
            if dry_run:
                continue

            raise ValueError('Cannot find any PostgreSQL databases to dump.')

        for database_name in dump_database_names:
            dumps_metadata.append(
                borgmatic.actions.restore.Dump(
                    'postgresql_databases',
                    database_name,
                    database.get('hostname'),
                    database.get('port'),
                    database.get('label'),
                    database.get('container'),
                )
            )
            dump_format = database.get('format', None if database_name == 'all' else 'custom')
            compression = database.get('compression')
            default_dump_command = 'pg_dumpall' if database_name == 'all' else 'pg_dump'
            dump_command = tuple(
                shlex.quote(part)
                for part in shlex.split(database.get('pg_dump_command') or default_dump_command)
            )
            dump_filename = dump.make_data_source_dump_filename(
                dump_path,
                database_name,
                hostname=database.get('hostname'),
                port=database.get('port'),
                container=database.get('container'),
                label=database.get('label'),
            )

            if os.path.exists(dump_filename):
                logger.warning(
                    f'Skipping duplicate dump of PostgreSQL database "{database_name}" to {dump_filename}',
                )
                continue

            hostname = database_config.resolve_database_option('hostname', database)
            command = (
                dump_command
                + (
                    '--no-password',
                    '--clean',
                    '--if-exists',
                )
                + (('--host', shlex.quote(hostname)) if hostname else ())
                + (('--port', shlex.quote(str(database['port']))) if 'port' in database else ())
                + (
                    (
                        '--username',
                        shlex.quote(
                            borgmatic.hooks.credential.parse.resolve_credential(
                                database['username'],
                                config,
                            ),
                        ),
                    )
                    if 'username' in database
                    else ()
                )
                + (('--no-owner',) if database.get('no_owner', False) else ())
                + (('--format', shlex.quote(dump_format)) if dump_format else ())
                + (('--compress', shlex.quote(str(compression))) if compression is not None else ())
                + (('--file', shlex.quote(dump_filename)) if dump_format == 'directory' else ())
                + (
                    tuple(shlex.quote(part) for part in shlex.split(database['options']))
                    if 'options' in database
                    else ()
                )
                + (() if database_name == 'all' else (shlex.quote(database_name),))
                # Use shell redirection rather than the --file flag to sidestep synchronization issues
                # when pg_dump/pg_dumpall tries to write to a named pipe. But for the directory dump
                # format in a particular, a named destination is required, and redirection doesn't work.
                + (('>', shlex.quote(dump_filename)) if dump_format != 'directory' else ())
            )

            logger.debug(
                f'Dumping PostgreSQL database "{database_name}" to {dump_filename}{dry_run_label}',
            )
            if dry_run:
                continue

            if dump_format == 'directory':
                dump.create_parent_directory_for_dump(dump_filename)
                execute_command(  # noqa: S604
                    command,
                    shell=True,
                    environment=environment,
                    working_directory=borgmatic.config.paths.get_working_directory(config),
                )
            else:
                dump.create_named_pipe_for_dump(dump_filename)
                processes.append(
                    execute_command(  # noqa: S604
                        command,
                        shell=True,
                        environment=environment,
                        run_to_completion=False,
                        working_directory=borgmatic.config.paths.get_working_directory(config),
                    ),
                )

    if not dry_run:
        dump.write_data_source_dumps_metadata(
            borgmatic_runtime_directory, 'postgresql_databases', dumps_metadata
        )
        borgmatic.hooks.data_source.config.inject_pattern(
            patterns,
            borgmatic.borg.pattern.Pattern(
                os.path.join(borgmatic_runtime_directory, 'postgresql_databases'),
                source=borgmatic.borg.pattern.Pattern_source.HOOK,
            ),
        )

    return processes


def remove_data_source_dumps(
    databases,
    config,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):  # pragma: no cover
    '''
    Remove all database dump files for this hook regardless of the given databases. Use the
    borgmatic runtime directory to construct the destination path. If this is a dry run, then don't
    actually remove anything.
    '''
    dump.remove_data_source_dumps(
        make_dump_path(borgmatic_runtime_directory),
        'PostgreSQL',
        dry_run,
    )


def make_data_source_dump_patterns(
    databases,
    config,
    borgmatic_runtime_directory,
    name=None,
    hostname=None,
    port=None,
    container=None,
    label=None,
):
    '''
    Given a sequence of configurations dicts, a configuration dict, the borgmatic runtime directory,
    and a database name to match, return the corresponding glob patterns to match the database dump
    in an archive.
    '''
    borgmatic_source_directory = borgmatic.config.paths.get_borgmatic_source_directory(config)

    return (
        *(
            dump.make_data_source_dump_filename(
                make_dump_path('borgmatic'), name, hostname, port, container, label
            ),
            dump.make_data_source_dump_filename(
                make_dump_path(borgmatic_runtime_directory),
                name,
                hostname,
                port,
                container,
                label,
            ),
            dump.make_data_source_dump_filename(
                make_dump_path(borgmatic_source_directory),
                name,
                hostname,
                port,
                container,
                label,
            ),
        ),
        *(
            (
                dump.make_data_source_dump_filename(
                    make_dump_path('borgmatic'),
                    name,
                    hostname,
                    port=None,
                    container=container,
                    label=label,
                ),
            )
            if port == get_default_port(databases, config)
            else ()
        ),
        *(
            (
                dump.make_data_source_dump_filename(
                    make_dump_path('borgmatic'),
                    name,
                    hostname,
                    port=get_default_port(databases, config),
                    container=container,
                    label=label,
                ),
            )
            if port is None
            else ()
        ),
    )


def restore_data_source_dump(
    hook_config,
    config,
    data_source,
    dry_run,
    extract_process,
    connection_params,
    borgmatic_runtime_directory,
):
    '''
    Restore a database from the given extract stream. The database is supplied as a data source
    configuration dict, but the given hook configuration is ignored. The given borgmatic runtime
    directory is used to construct the destination path (used for the directory format). If this is
    a dry run, then don't actually restore anything. Trigger the given active extract process (an
    instance of subprocess.Popen) to produce output to consume.

    If the extract process is None, then restore the dump from the filesystem rather than from an
    extract stream.

    Use the given connection parameters to connect to the database. The connection parameters are
    hostname, port, username, and password.
    '''
    dry_run_label = ' (dry run; not actually restoring anything)' if dry_run else ''
    hostname = database_config.resolve_database_option(
        'hostname', data_source, connection_params, restore=True
    )
    port = database_config.resolve_database_option(
        'port', data_source, connection_params, restore=True
    )
    username = borgmatic.hooks.credential.parse.resolve_credential(
        database_config.resolve_database_option(
            'username', data_source, connection_params, restore=True
        ),
        config,
    )

    all_databases = bool(data_source['name'] == 'all')
    dump_filename = dump.make_data_source_dump_filename(
        make_dump_path(borgmatic_runtime_directory),
        data_source['name'],
        hostname=hostname,
        port=port,
        container=data_source.get('container'),
        label=data_source.get('label'),
    )
    psql_command = tuple(
        shlex.quote(part) for part in shlex.split(data_source.get('psql_command') or 'psql')
    )
    analyze_command = (
        psql_command
        + ('--no-password', '--no-psqlrc', '--quiet')
        + (('--host', hostname) if hostname else ())
        + (('--port', str(port)) if port else ())
        + (('--username', username) if username else ())
        + (('--dbname', data_source['name']) if not all_databases else ())
        + (
            tuple(shlex.quote(part) for part in shlex.split(data_source['analyze_options']))
            if 'analyze_options' in data_source
            else ()
        )
        + ('--command', 'ANALYZE')
    )
    use_psql_command = all_databases or data_source.get('format') == 'plain'
    pg_restore_command = tuple(
        shlex.quote(part)
        for part in shlex.split(data_source.get('pg_restore_command') or 'pg_restore')
    )
    restore_command = (
        (psql_command if use_psql_command else pg_restore_command)
        + ('--no-password',)
        + (('--no-psqlrc',) if use_psql_command else ('--if-exists', '--exit-on-error', '--clean'))
        + (('--dbname', data_source['name']) if not all_databases else ())
        + (('--host', hostname) if hostname else ())
        + (('--port', str(port)) if port else ())
        + (('--username', username) if username else ())
        + (('--no-owner',) if data_source.get('no_owner', False) else ())
        + (
            tuple(shlex.quote(part) for part in shlex.split(data_source['restore_options']))
            if 'restore_options' in data_source
            else ()
        )
        + (() if extract_process else (str(pathlib.Path(dump_filename)),))
        + tuple(
            itertools.chain.from_iterable(('--schema', schema) for schema in data_source['schemas'])
            if data_source.get('schemas')
            else (),
        )
    )

    environment = make_environment(data_source, config, restore_connection_params=connection_params)

    logger.debug(f"Restoring PostgreSQL database {data_source['name']}{dry_run_label}")
    if dry_run:
        return

    # Don't give Borg local path so as to error on warnings, as "borg extract" only gives a warning
    # if the restore paths don't exist in the archive.
    tuple(
        execute_command_with_processes(
            restore_command,
            [extract_process] if extract_process else [],
            output_log_level=logging.DEBUG,
            input_file=extract_process.stdout if extract_process else None,
            environment=environment,
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=config.get('local_path', 'borg'),
        )
    )
    execute_command(
        analyze_command,
        environment=environment,
        working_directory=borgmatic.config.paths.get_working_directory(config),
    )



import enum
import json
import logging
import logging.handlers
import os
import socket
import sys


def to_bool(arg):
    '''
    Return a boolean value based on `arg`.
    '''
    if arg is None or isinstance(arg, bool):
        return arg

    if isinstance(arg, str):
        arg = arg.lower()

    return arg in {'yes', 'on', '1', 'true', 1}


def interactive_console():
    '''
    Return whether the current console is "interactive". Meaning: Capable of
    user input and not just something like a cron job.
    '''
    return sys.stderr.isatty() and os.environ.get('TERM') != 'dumb'


def should_do_markup(configs, json_enabled):
    '''
    Given a dict of configuration filename to corresponding parsed configuration (which already have
    any command-line overrides applied) and whether json is enabled, determine if we should enable
    color marking up.
    '''
    if json_enabled:
        return False

    if any(config.get('color', True) is False for config in configs.values()):
        return False

    if os.environ.get('NO_COLOR', None):
        return False

    py_colors = os.environ.get('PY_COLORS', None)

    if py_colors is not None:
        return to_bool(py_colors)

    return interactive_console()


class Multi_stream_handler(logging.Handler):
    '''
    A logging handler that dispatches each log record to one of multiple stream handlers depending
    on the record's log level.
    '''

    def __init__(self, log_level_to_stream_handler):
        super().__init__()
        self.log_level_to_handler = log_level_to_stream_handler
        self.handlers = set(self.log_level_to_handler.values())

    def flush(self):  # pragma: no cover
        super().flush()

        for handler in self.handlers:
            handler.flush()

    def emit(self, record):
        '''
        Dispatch the log record to the appropriate stream handler for the record's log level.
        '''
        self.log_level_to_handler[record.levelno].emit(record)

    def setFormatter(self, formatter):  # pragma: no cover  # noqa: N802
        super().setFormatter(formatter)

        for handler in self.handlers:
            handler.setFormatter(formatter)

    def setLevel(self, level):  # pragma: no cover  # noqa: N802
        super().setLevel(level)

        for handler in self.handlers:
            handler.setLevel(level)


DEFAULT_JOURNALD_PRIORITY = 6


class JournaldHandler(logging.Handler):
    def __init__(self, journald_socket_path):
        super().__init__()

        add_custom_log_levels()

        self.journald_socket_path = journald_socket_path
        self.log_level_to_journald_priority = {
            logging.CRITICAL: 2,
            logging.ERROR: 3,
            logging.WARNING: 4,
            logging.ANSWER: 5,
            logging.INFO: 6,
            logging.DEBUG: 7,
        }

    def emit(self, record):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        try:
            message_parts = []
            entry = dict(
                MESSAGE=record.getMessage(),
                PRIORITY=self.log_level_to_journald_priority.get(
                    record.levelno, DEFAULT_JOURNALD_PRIORITY
                ),
                SYSLOG_IDENTIFIER='borgmatic',
                SYSLOG_PID=os.getpid(),
            )

            for key, value in entry.items():
                encoded_key = key.upper().encode('utf-8')
                encoded_value = str(value).encode('utf-8')

                # Multi-line and single-line values use different formats on the wire.
                if b'\n' in encoded_value:
                    message_parts.extend((encoded_key, b'\n'))
                    message_parts.extend(
                        (len(encoded_value).to_bytes(8, 'little'), encoded_value, b'\n')
                    )
                else:
                    message_parts.extend((encoded_key, b'=', encoded_value, b'\n'))

            sock.sendto(b''.join(message_parts), self.journald_socket_path)
        finally:
            sock.close()


class Log_prefix_formatter(logging.Formatter):
    def __init__(self, fmt='{prefix}{message}', *args, style='{', **kwargs):
        self.prefix = None

        super().__init__(*args, fmt=fmt, style=style, **kwargs)

    def format(self, record):  # pragma: no cover
        record.prefix = f'{self.prefix}: ' if self.prefix else ''

        return super().format(record)


def log_record_to_json(record):
    '''
    Given a logging.LogRecord, return it as a JSON-encoded string containing relevant attributes.
    '''
    message_id = getattr(record, 'msgid', None)

    return json.dumps(
        dict(
            type='log_message',
            time=record.created,
            message=record.getMessage(),
            levelname=record.levelname,
            name=record.name,
        )
        | ({'msgid': message_id} if message_id is not None else {})
    )


class Json_formatter(logging.Formatter):
    def __init__(self, fmt='{message}', *args, style='{', **kwargs):
        super().__init__(*args, fmt=fmt, style=style, **kwargs)

    def format(self, record):
        return log_record_to_json(record)


class Color(enum.Enum):
    RESET = 0
    RED = 31
    GREEN = 32
    YELLOW = 33
    MAGENTA = 35
    CYAN = 36


class Console_color_formatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self.prefix = None
        super().__init__(
            '{prefix}{message}',
            *args,
            style='{',
            **kwargs,
        )

    def format(self, record):
        add_custom_log_levels()

        color = {
            logging.CRITICAL: Color.RED,
            logging.ERROR: Color.RED,
            logging.WARNING: Color.YELLOW,
            logging.ANSWER: Color.MAGENTA,
            logging.INFO: Color.GREEN,
            logging.DEBUG: Color.CYAN,
        }.get(record.levelno).value
        record.prefix = f'{self.prefix}: ' if self.prefix else ''

        return color_text(color, super().format(record))


def ansi_escape_code(color):  # pragma: no cover
    '''
    Given a color value, produce the corresponding ANSI escape code.
    '''
    return f'\x1b[{color}m'


def color_text(color, message):
    '''
    Given a color value and a message, return the message colored with ANSI escape codes.
    '''
    if not color:
        return message

    return f'{ansi_escape_code(color)}{message}{ansi_escape_code(Color.RESET.value)}'


def add_logging_level(level_name, level_number):
    '''
    Globally add a custom logging level based on the given (all uppercase) level name and number.
    Do this idempotently.

    Inspired by https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
    '''
    method_name = level_name.lower()

    if not hasattr(logging, level_name):
        logging.addLevelName(level_number, level_name)
        setattr(logging, level_name, level_number)

    if not hasattr(logging, method_name):

        def log_for_level(self, message, *args, **kwargs):  # pragma: no cover
            if self.isEnabledFor(level_number):
                self._log(level_number, message, args, **kwargs)

        setattr(logging.getLoggerClass(), method_name, log_for_level)

    if not hasattr(logging.getLoggerClass(), method_name):

        def log_to_root(message, *args, **kwargs):  # pragma: no cover
            logging.log(level_number, message, *args, **kwargs)  # noqa: LOG015

        setattr(logging, method_name, log_to_root)


ANSWER = logging.WARNING - 5
DISABLED = logging.CRITICAL + 10


def add_custom_log_levels():  # pragma: no cover
    '''
    Add a custom log level between WARNING and INFO for user-requested answers.
    '''
    add_logging_level('ANSWER', ANSWER)
    add_logging_level('DISABLED', DISABLED)


def get_log_prefix():
    '''
    Return the current log prefix set by set_log_prefix(). Return None if no such prefix exists.

    It would be a whole lot easier to use logger.Formatter(defaults=...) instead, but that argument
    doesn't exist until Python 3.10+.
    '''
    try:
        formatter = next(
            handler.formatter
            for handler in logging.getLogger().handlers
            if handler.formatter
            if hasattr(handler.formatter, 'prefix')
        )
    except StopIteration:
        return None

    return formatter.prefix


def set_log_prefix(prefix):
    '''
    Given a log prefix as a string, set it into the each handler's formatter so that it can inject
    the prefix into each logged record.
    '''
    for handler in logging.getLogger().handlers:
        if handler.formatter and hasattr(handler.formatter, 'prefix'):
            handler.formatter.prefix = prefix


class Log_prefix:
    '''
    A Python context manager for setting a log prefix so that it shows up in every subsequent
    logging message for the duration of the context manager. For this to work, it relies on each
    logging formatter to be initialized with "{prefix}" somewhere in its logging format.

    Example use as a context manager:


       with borgmatic.logger.Log_prefix('myprefix'):
            do_something_that_logs()

    For the scope of that "with" statement, any logs created are prefixed with "myprefix: ".
    Afterwards, the prefix gets restored to whatever it was prior to the context manager.
    '''

    def __init__(self, prefix):
        '''
        Given the desired log prefix, save it for use below. Set prefix to None to disable any
        prefix from getting logged.
        '''
        self.prefix = prefix
        self.original_prefix = None

    def __enter__(self):
        '''
        Set the prefix onto the formatter defaults for every logging handler so that the prefix ends
        up in every log message. But first, save off any original prefix so that it can be restored
        below.
        '''
        self.original_prefix = get_log_prefix()
        set_log_prefix(self.prefix)

    def __exit__(self, exception_type, exception, traceback):
        '''
        Restore any original prefix.
        '''
        set_log_prefix(self.original_prefix)


class Log_exclude_filter(logging.Filter):
    '''
    A Python log filter that omits log records matching given attributes.
    '''

    def __init__(self, name, filter_attributes):
        '''
        Given a unique name for this filter and a dict of attributes to filter on, set the filter
        name and save the attributes for use below.
        '''
        self.filter_attributes = filter_attributes

        super().__init__(name)

    def filter(self, log_record):
        '''
        Given a log record, return False (indicating the record should be omitted) if the record's
        attributes match any of the saved filter attributes. Return True (indicating do not omit)
        otherwise.
        '''
        for attribute_name, value in self.filter_attributes.items():
            if getattr(log_record, attribute_name, None) == value:
                return False

        return True


def add_log_exclude_filter(name, filter_attributes):
    '''
    Given a unique filter name and a dict of attributes to filter on, create a log exclude filter
    with them and add the filter to each log handler.
    '''
    for handler in logging.getLogger().handlers:
        handler.addFilter(Log_exclude_filter(name, filter_attributes))


def remove_log_exclude_filter(name):
    '''
    Given a unique filter name, remove matching filters from each log handler.
    '''
    for handler in logging.getLogger().handlers:
        for exclude_filter in handler.filters:
            if getattr(exclude_filter, 'name', None) == name:
                handler.removeFilter(exclude_filter)


class Logs_suppressed:
    '''
    A Python context manager for temporarily adding a log filter that suppresses requested log
    records for the duration of the context manager.

    Example use:


       with borgmatic.logger.Logs_suppressed(msgid='Repository.DoesNotExist'):
            do_something_that_logs()

    For the scope of that "with" statement, any records logged with the given message ID are
    filtered out of the log output. "msgid" is just an example; any logging.LogRecord attributes
    (standard or custom) can be passed in to filter on.

    Multiple instances of this context manager with different filter attributes can be in use at
    once.
    '''

    def __init__(self, **filter_attributes):
        '''
        Given the desired log record filter attributes as keyword arguments, save them for use below.
        '''
        self.filter_attributes = filter_attributes

    def __enter__(self):
        '''
        Create a log filter with the saved filter attributes and add the filter to every logging
        handler, so that they filter out the desired log records.
        '''
        add_log_exclude_filter(name=str(id(self)), filter_attributes=self.filter_attributes)

    def __exit__(self, exception_type, exception, traceback):
        '''
        Remove the previously added filter from every logging handler.
        '''
        remove_log_exclude_filter(name=str(id(self)))


class Delayed_logging_handler(logging.handlers.BufferingHandler):
    '''
    A logging handler that buffers logs and doesn't flush them until explicitly flushed (after
    target handlers are actually set). It's useful for holding onto messages logged before logging
    is configured, ensuring those records eventually make their way to the relevant logging
    handlers.

    When flushing, don't forward log records to a target handler if the record's log level is below
    that of the handler. This recreates the standard logging behavior of, say, logging.DEBUG records
    getting suppressed if a handler's level is only set to logging.INFO.
    '''

    def __init__(self):
        super().__init__(capacity=0)

        self.targets = None

    def shouldFlush(self, record):  # noqa: N802
        return self.targets is not None

    def flush(self):
        self.acquire()

        try:
            if not self.targets:
                return

            for record in self.buffer:
                for target in self.targets:
                    if record.levelno >= target.level:
                        target.handle(record)

            self.buffer.clear()
        finally:
            self.release()


def configure_delayed_logging():  # pragma: no cover
    '''
    Configure a delayed logging handler to buffer anything that gets logged until we're ready to
    deal with it.
    '''
    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[Delayed_logging_handler()],
    )


def flush_delayed_logging(target_handlers):
    '''
    Flush any previously buffered logs to our "real" logging handlers.
    '''
    root_logger = logging.getLogger()

    if root_logger.handlers and isinstance(root_logger.handlers[0], Delayed_logging_handler):
        delayed_handler = root_logger.handlers[0]
        delayed_handler.targets = target_handlers
        delayed_handler.flush()
        root_logger.removeHandler(delayed_handler)


JOURNALD_SOCKET_PATH = '/run/systemd/journal/socket'
SYSLOG_PATHS = ('/dev/log', '/var/run/syslog', '/var/run/log')


def configure_logging(
    console_log_level,
    syslog_log_level=None,
    log_file_log_level=None,
    monitoring_log_level=None,
    log_file=None,
    log_file_format=None,
    log_json=False,
    color_enabled=True,
):
    '''
    Configure logging to go to both the console and (syslog or log file). Use the given log levels,
    respectively. If color is enabled, set up log formatting accordingly.

    Raise FileNotFoundError or PermissionError if the log file could not be opened for writing.
    '''
    add_custom_log_levels()

    if syslog_log_level is None:
        syslog_log_level = logging.DISABLED

    if log_file_log_level is None:
        log_file_log_level = console_log_level

    if monitoring_log_level is None:
        monitoring_log_level = console_log_level

    # Log certain log levels to console stderr and others to stdout. This supports use cases like
    # grepping (non-error) output.
    console_disabled = logging.NullHandler()
    console_error_handler = logging.StreamHandler(sys.stderr)
    console_standard_handler = logging.StreamHandler(sys.stdout)
    console_handler = Multi_stream_handler(
        {
            logging.DISABLED: console_disabled,
            logging.CRITICAL: console_error_handler,
            logging.ERROR: console_error_handler,
            logging.WARNING: console_error_handler,
            logging.ANSWER: console_standard_handler,
            logging.INFO: console_standard_handler,
            logging.DEBUG: console_standard_handler,
        },
    )

    if log_json:
        console_handler.setFormatter(Json_formatter())
    elif color_enabled:
        console_handler.setFormatter(Console_color_formatter())
    else:
        console_handler.setFormatter(Log_prefix_formatter())

    console_handler.setLevel(console_log_level)
    handlers = [console_handler]

    if syslog_log_level != logging.DISABLED:
        if os.path.exists(JOURNALD_SOCKET_PATH):
            journald_handler = JournaldHandler(JOURNALD_SOCKET_PATH)
            journald_handler.setLevel(syslog_log_level)
            handlers.append(journald_handler)
        else:
            syslog_path = next(
                (path for path in SYSLOG_PATHS if os.path.exists(path)),
                None,
            )

            if syslog_path:
                syslog_handler = logging.handlers.SysLogHandler(address=syslog_path)
                syslog_handler.setFormatter(
                    Log_prefix_formatter(
                        'borgmatic: {levelname} {prefix}{message}',
                    ),
                )
                syslog_handler.setLevel(syslog_log_level)
                handlers.append(syslog_handler)

    if log_file and log_file_log_level != logging.DISABLED:
        file_handler = logging.handlers.WatchedFileHandler(log_file)
        file_handler.setFormatter(
            Json_formatter()
            if log_json
            else Log_prefix_formatter(
                log_file_format or '[{asctime}] {levelname}: {prefix}{message}',
            ),
        )
        file_handler.setLevel(log_file_log_level)
        handlers.append(file_handler)

    flush_delayed_logging(handlers)

    logging.basicConfig(
        level=min(handler.level for handler in handlers),
        handlers=handlers,
    )



import logging
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, flags
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


def export_tar_archive(
    dry_run,
    repository_path,
    archive,
    paths,
    destination_path,
    config,
    local_borg_version,
    global_arguments,
    local_path='borg',
    remote_path=None,
    tar_filter=None,
    strip_components=None,
):
    '''
    Given a dry-run flag, a local or remote repository path, an archive name, zero or more paths to
    export from the archive, a destination path to export to, a configuration dict, the local Borg
    version, optional local and remote Borg paths, an optional filter program, whether to include
    per-file details, and an optional number of path components to strip, export the archive into
    the given destination path as a tar-formatted file.

    If the destination path is "-", then stream the output to stdout instead of to a file.
    '''
    borgmatic.logger.add_custom_log_levels()
    umask = config.get('umask')
    lock_wait = config.get('lock_wait')
    extra_borg_options = config.get('extra_borg_options', {}).get('export_tar', '')

    full_command = (
        (local_path, 'export-tar')
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (('--log-json',) if destination_path != '-' else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--list',) if config.get('list_details') else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (('--dry-run',) if dry_run else ())
        + (('--tar-filter', tar_filter) if tar_filter else ())
        + (('--strip-components', str(strip_components)) if strip_components else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_archive_flags(
            repository_path,
            archive,
            local_borg_version,
        )
        + (destination_path,)
        + (tuple(paths) if paths else ())
    )

    output_log_level = logging.ANSWER if config.get('list_details') else logging.INFO

    if dry_run:
        logger.info('Skipping export to tar file (dry run)')
        return

    execute_command(
        full_command,
        output_file=DO_NOT_CAPTURE if destination_path == '-' else None,
        output_log_level=output_log_level,
        environment=environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import textual.widgets

import borgmatic.actions.browse.bindings


class Repositories_list(textual.widgets.OptionList):
    '''
    A widget for selecting a single Borg repository from among the repositories in a borgmatic
    configuration file. The item selection event is handled in a Carousel instance, the parent
    widget of a Repositories_list.
    '''

    BINDINGS = borgmatic.actions.browse.bindings.OPTION_LIST_BINDINGS

    def __init__(self, config):
        '''
        Given a configuration dict, populate the repositories in this widget.
        '''
        self.config = config
        self.repositories = config['repositories']

        super().__init__(
            *(
                textual.widgets.option_list.Option(label, id=index)
                for index, repository in enumerate(self.repositories)
                for label in (repository.get('label', repository.get('path')),)
            ),
            classes='panel',
        )
        self.border_title = '📦 repositories'



import logging
import os
import shlex

import borgmatic.borg.pattern
import borgmatic.config.paths
import borgmatic.hooks.credential.parse
import borgmatic.hooks.data_source.config
import borgmatic.hooks.data_source.mariadb
from borgmatic.execute import (
    execute_command,
    execute_command_and_capture_output,
    execute_command_with_processes,
)
from borgmatic.hooks.data_source import config as database_config
from borgmatic.hooks.data_source import dump

logger = logging.getLogger(__name__)


def make_dump_path(base_directory):  # pragma: no cover
    '''
    Given a base directory, make the corresponding dump path.
    '''
    return dump.make_data_source_dump_path(base_directory, 'mysql_databases')


SYSTEM_DATABASE_NAMES = ('information_schema', 'mysql', 'performance_schema', 'sys')


def database_names_to_dump(database, config, username, password, environment, dry_run):
    '''
    Given a requested database config, a configuration dict, a database username and password, an
    environment dict, and whether this is a dry run, return the corresponding sequence of database
    names to dump. In the case of "all", query for the names of databases on the configured host and
    return them, excluding any system databases that will cause problems during restore.
    '''
    skip_names = database.get('skip_names')

    if database['name'] != 'all':
        if skip_names:
            logger.warning(
                f'For MySQL database {database["name"]}, ignoring the "skip_names" option, which is only supported for database "all"'
            )

        return (database['name'],)

    if dry_run:
        return ()

    mysql_show_command = tuple(
        shlex.quote(part) for part in shlex.split(database.get('mysql_command') or 'mysql')
    )
    extra_options, defaults_extra_filename = (
        borgmatic.hooks.data_source.mariadb.parse_extra_options(database.get('list_options'))
    )
    password_transport = database.get('password_transport', 'pipe')
    hostname = database_config.resolve_database_option('hostname', database)
    socket_path = database.get('socket_path')

    show_command = (
        mysql_show_command
        + (
            borgmatic.hooks.data_source.mariadb.make_defaults_file_options(
                username,
                password,
                defaults_extra_filename,
            )
            if password_transport == 'pipe'
            else ()
        )
        + extra_options
        + (('--host', hostname) if hostname else ())
        + (('--port', str(database['port'])) if 'port' in database else ())
        + (('--protocol', 'tcp') if hostname or 'port' in database else ())
        + (('--socket', socket_path) if socket_path else ())
        + (('--user', username) if username and password_transport == 'environment' else ())
        + (('--ssl',) if database.get('tls') is True else ())
        + (('--skip-ssl',) if database.get('tls') is False else ())
        + ('--skip-column-names', '--batch')
        + ('--execute', 'show schemas')
    )

    logger.debug('Querying for "all" MySQL databases to dump')

    if skip_names:
        logger.debug(f'Skipping database names: {", ".join(skip_names)}')

    show_lines = execute_command_and_capture_output(
        show_command,
        environment=environment,
        working_directory=borgmatic.config.paths.get_working_directory(config),
    )

    return tuple(
        show_name.strip()
        for show_name in show_lines
        if show_name not in SYSTEM_DATABASE_NAMES
        if not skip_names or show_name not in skip_names
    )


def execute_dump_command(
    database,
    config,
    username,
    password,
    dump_path,
    database_names,
    environment,
    dry_run,
    dry_run_label,
):
    '''
    Kick off a dump for the given MySQL/MariaDB database (provided as a configuration dict) to a
    named pipe constructed from the given dump path and database name.

    Return a subprocess.Popen instance for the dump process ready to spew to a named pipe. But if
    this is a dry run, then don't actually dump anything and return None.
    '''
    database_name = database['name']
    dump_filename = dump.make_data_source_dump_filename(
        dump_path,
        database['name'],
        hostname=database.get('hostname'),
        port=database.get('port'),
        container=database.get('container'),
        label=database.get('label'),
    )

    if os.path.exists(dump_filename):
        logger.warning(
            f'Skipping duplicate dump of MySQL database "{database_name}" to {dump_filename}',
        )
        return None

    mysql_dump_command = tuple(
        shlex.quote(part) for part in shlex.split(database.get('mysql_dump_command') or 'mysqldump')
    )
    extra_options, defaults_extra_filename = (
        borgmatic.hooks.data_source.mariadb.parse_extra_options(database.get('options'))
    )
    password_transport = database.get('password_transport', 'pipe')
    hostname = database_config.resolve_database_option('hostname', database)
    socket_path = database.get('socket_path')

    dump_command = (
        mysql_dump_command
        + (
            borgmatic.hooks.data_source.mariadb.make_defaults_file_options(
                username,
                password,
                defaults_extra_filename,
            )
            if password_transport == 'pipe'
            else ()
        )
        + extra_options
        + (('--add-drop-database',) if database.get('add_drop_database', True) else ())
        + ('--single-transaction',)
        + (('--host', hostname) if hostname else ())
        + (('--port', str(database['port'])) if 'port' in database else ())
        + (('--protocol', 'tcp') if hostname or 'port' in database else ())
        + (('--socket', socket_path) if socket_path else ())
        + (('--user', username) if username and password_transport == 'environment' else ())
        + (('--ssl',) if database.get('tls') is True else ())
        + (('--skip-ssl',) if database.get('tls') is False else ())
        + (('--events',) if database.get('events', True) else ())
        + (('--routines',) if database.get('routines', True) else ())
        + (('--all-tablespaces',) if database.get('tablespaces', True) else ())
        + ('--databases',)
        + database_names
        + ('--result-file', dump_filename)
    )

    logger.debug(f'Dumping MySQL database "{database_name}" to {dump_filename}{dry_run_label}')
    if dry_run:
        return None

    dump.create_named_pipe_for_dump(dump_filename)

    return execute_command(
        dump_command,
        environment=environment,
        run_to_completion=False,
        working_directory=borgmatic.config.paths.get_working_directory(config),
    )


def get_default_port(databases, config):  # pragma: no cover
    return 3306


def use_streaming(databases, config):
    '''
    Given a sequence of MySQL database configuration dicts, a configuration dict (ignored), return
    whether streaming will be using during dumps.
    '''
    return any(databases)


def dump_data_sources(
    databases,
    config,
    config_paths,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):
    '''
    Dump the given MySQL/MariaDB databases to a named pipe. The databases are supplied as a sequence
    of dicts, one dict describing each database as per the configuration schema. Use the given
    borgmatic runtime directory to construct the destination path.

    Return a sequence of subprocess.Popen instances for the dump processes ready to spew to a named
    pipe. But if this is a dry run, then don't actually dump anything and return an empty sequence.
    Also append the the parent directory of the database dumps to the given patterns list, so the
    dumps actually get backed up.
    '''
    dry_run_label = ' (dry run; not actually dumping anything)' if dry_run else ''
    processes = []
    dumps_metadata = []

    logger.info(f'Dumping MySQL databases{dry_run_label}')
    dump_path = make_dump_path(borgmatic_runtime_directory)

    for database in databases:
        username = borgmatic.hooks.credential.parse.resolve_credential(
            database.get('username'),
            config,
        )
        password = borgmatic.hooks.credential.parse.resolve_credential(
            database.get('password'),
            config,
        )
        environment = dict(
            os.environ,
            **(
                {'MYSQL_PWD': password}
                if password and database.get('password_transport') == 'environment'
                else {}
            ),
        )
        dump_database_names = database_names_to_dump(
            database,
            config,
            username,
            password,
            environment,
            dry_run,
        )

        if not dump_database_names:
            if dry_run:
                continue

            raise ValueError('Cannot find any MySQL databases to dump.')

        if database['name'] == 'all' and database.get('format'):
            for database_name in dump_database_names:
                dumps_metadata.append(
                    borgmatic.actions.restore.Dump(
                        'mysql_databases',
                        database_name,
                        database.get('hostname'),
                        database.get('port'),
                        database.get('label'),
                        database.get('container'),
                    )
                )
                renamed_database = dict(database, name=database_name)
                processes.append(
                    execute_dump_command(
                        renamed_database,
                        config,
                        username,
                        password,
                        dump_path,
                        (database_name,),
                        environment,
                        dry_run,
                        dry_run_label,
                    ),
                )
        else:
            dumps_metadata.append(
                borgmatic.actions.restore.Dump(
                    'mysql_databases',
                    database['name'],
                    database.get('hostname'),
                    database.get('port'),
                    database.get('label'),
                    database.get('container'),
                )
            )
            processes.append(
                execute_dump_command(
                    database,
                    config,
                    username,
                    password,
                    dump_path,
                    dump_database_names,
                    environment,
                    dry_run,
                    dry_run_label,
                ),
            )

    if not dry_run:
        dump.write_data_source_dumps_metadata(
            borgmatic_runtime_directory, 'mysql_databases', dumps_metadata
        )
        borgmatic.hooks.data_source.config.inject_pattern(
            patterns,
            borgmatic.borg.pattern.Pattern(
                os.path.join(borgmatic_runtime_directory, 'mysql_databases'),
                source=borgmatic.borg.pattern.Pattern_source.HOOK,
            ),
        )

    return [process for process in processes if process]


def remove_data_source_dumps(
    databases,
    config,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):  # pragma: no cover
    '''
    Remove all database dump files for this hook regardless of the given databases. Use the
    borgmatic runtime directory to construct the destination path. If this is a dry run, then don't
    actually remove anything.
    '''
    dump.remove_data_source_dumps(make_dump_path(borgmatic_runtime_directory), 'MySQL', dry_run)


def make_data_source_dump_patterns(
    databases,
    config,
    borgmatic_runtime_directory,
    name=None,
    hostname=None,
    port=None,
    container=None,
    label=None,
):
    '''
    Given a sequence of configurations dicts, a configuration dict, the borgmatic runtime directory,
    and a database name to match, return the corresponding glob patterns to match the database dump
    in an archive.
    '''
    borgmatic_source_directory = borgmatic.config.paths.get_borgmatic_source_directory(config)

    return (
        *(
            dump.make_data_source_dump_filename(
                make_dump_path('borgmatic'), name, hostname, port, container, label
            ),
            dump.make_data_source_dump_filename(
                make_dump_path(borgmatic_runtime_directory),
                name,
                hostname,
                port,
                container,
                label,
            ),
            dump.make_data_source_dump_filename(
                make_dump_path(borgmatic_source_directory),
                name,
                hostname,
                port,
                container,
                label,
            ),
        ),
        *(
            (
                dump.make_data_source_dump_filename(
                    make_dump_path('borgmatic'),
                    name,
                    hostname,
                    port=None,
                    container=container,
                    label=label,
                ),
            )
            if port == get_default_port(databases, config)
            else ()
        ),
        *(
            (
                dump.make_data_source_dump_filename(
                    make_dump_path('borgmatic'),
                    name,
                    hostname,
                    port=get_default_port(databases, config),
                    container=container,
                    label=label,
                ),
            )
            if port is None
            else ()
        ),
    )


def restore_data_source_dump(
    hook_config,
    config,
    data_source,
    dry_run,
    extract_process,
    connection_params,
    borgmatic_runtime_directory,
):
    '''
    Restore a database from the given extract stream. The database is supplied as a data source
    configuration dict, but the given hook configuration is ignored. If this is a dry run, then
    don't actually restore anything. Trigger the given active extract process (an instance of
    subprocess.Popen) to produce output to consume.
    '''
    dry_run_label = ' (dry run; not actually restoring anything)' if dry_run else ''
    hostname = database_config.resolve_database_option(
        'hostname', data_source, connection_params, restore=True
    )
    port = database_config.resolve_database_option(
        'port', data_source, connection_params, restore=True
    )
    socket_path = database_config.resolve_database_option('socket_path', data_source, restore=True)
    tls = database_config.resolve_database_option('tls', data_source, restore=True)
    username = borgmatic.hooks.credential.parse.resolve_credential(
        database_config.resolve_database_option(
            'username', data_source, connection_params, restore=True
        ),
        config,
    )
    password = borgmatic.hooks.credential.parse.resolve_credential(
        database_config.resolve_database_option(
            'password', data_source, connection_params, restore=True
        ),
        config,
    )

    mysql_restore_command = tuple(
        shlex.quote(part) for part in shlex.split(data_source.get('mysql_command') or 'mysql')
    )
    extra_options, defaults_extra_filename = (
        borgmatic.hooks.data_source.mariadb.parse_extra_options(data_source.get('restore_options'))
    )
    password_transport = data_source.get('password_transport', 'pipe')

    restore_command = (
        mysql_restore_command
        + (
            borgmatic.hooks.data_source.mariadb.make_defaults_file_options(
                username,
                password,
                defaults_extra_filename,
            )
            if password_transport == 'pipe'
            else ()
        )
        + extra_options
        + ('--batch',)
        + (('--host', hostname) if hostname else ())
        + (('--port', str(port)) if port else ())
        + (('--protocol', 'tcp') if hostname or port else ())
        + (('--socket', socket_path) if socket_path else ())
        + (('--user', username) if username and password_transport == 'environment' else ())
        + (('--ssl',) if tls is True else ())
        + (('--skip-ssl',) if tls is False else ())
    )
    environment = dict(
        os.environ,
        **({'MYSQL_PWD': password} if password and password_transport == 'environment' else {}),
    )

    logger.debug(f"Restoring MySQL database {data_source['name']}{dry_run_label}")
    if dry_run:
        return

    # Don't give Borg local path so as to error on warnings, as "borg extract" only gives a warning
    # if the restore paths don't exist in the archive.
    tuple(
        execute_command_with_processes(
            restore_command,
            [extract_process],
            output_log_level=logging.DEBUG,
            input_file=extract_process.stdout,
            environment=environment,
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=config.get('local_path', 'borg'),
        )
    )



import logging
import os


def normalize_sections(config_filename, config):
    '''
    Given a configuration filename and a configuration dict of its loaded contents, airlift any
    options out of sections ("location:", etc.) to the global scope and delete those sections.
    Return any log message warnings produced based on the normalization performed.

    Raise ValueError if the "prefix" option is set in both "location" and "consistency" sections.
    '''
    try:
        location = config.get('location') or {}
    except AttributeError:
        raise ValueError('Configuration does not contain any options')

    storage = config.get('storage') or {}
    consistency = config.get('consistency') or {}
    hooks = config.get('hooks') or {}

    if (
        location.get('prefix')
        and consistency.get('prefix')
        and location.get('prefix') != consistency.get('prefix')
    ):
        raise ValueError(
            'The retention prefix and the consistency prefix cannot have different values (unless one is not set).',
        )

    if storage.get('umask') and hooks.get('umask') and storage.get('umask') != hooks.get('umask'):
        raise ValueError(
            'The storage umask and the hooks umask cannot have different values (unless one is not set).',
        )

    any_section_upgraded = False

    # Move any options from deprecated sections into the global scope.
    for section_name in ('location', 'storage', 'retention', 'consistency', 'output', 'hooks'):
        section_config = config.get(section_name)

        if section_config is not None:
            any_section_upgraded = True
            del config[section_name]
            config.update(section_config)

    if any_section_upgraded:
        return [
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: Configuration sections (like location:, storage:, retention:, consistency:, and hooks:) are deprecated and support will be removed from a future release. To prepare for this, move your options out of sections to the global scope.',
                ),
            ),
        ]

    return []


def make_command_hook_deprecation_log(config_filename, option_name):  # pragma: no cover
    '''
    Given a configuration filename and the name of a configuration option, return a deprecation
    warning log for it.
    '''
    return logging.makeLogRecord(
        dict(
            levelno=logging.WARNING,
            levelname='WARNING',
            msg=f'{config_filename}: {option_name} is deprecated and support will be removed from a future release. Use commands: instead.',
        ),
    )


def normalize_commands(config_filename, config):
    '''
    Given a configuration filename and a configuration dict, transform any "before_*"- and
    "after_*"-style command hooks into "commands:".
    '''
    logs = []

    # Normalize "before_actions" and "after_actions".
    for preposition in ('before', 'after'):
        option_name = f'{preposition}_actions'
        commands = config.pop(option_name, None)

        if commands:
            logs.append(make_command_hook_deprecation_log(config_filename, option_name))
            config.setdefault('commands', []).append(
                {
                    preposition: 'repository',
                    'run': commands,
                },
            )

    # Normalize "before_backup", "before_prune", "after_backup", "after_prune", etc.
    for action_name in ('create', 'prune', 'compact', 'check', 'extract'):
        for preposition in ('before', 'after'):
            option_name = f'{preposition}_{"backup" if action_name == "create" else action_name}'
            commands = config.pop(option_name, None)

            if not commands:
                continue

            logs.append(make_command_hook_deprecation_log(config_filename, option_name))
            config.setdefault('commands', []).append(
                {
                    preposition: 'action',
                    'when': [action_name],
                    'run': commands,
                },
            )

    # Normalize "on_error".
    commands = config.pop('on_error', None)

    if commands:
        logs.append(make_command_hook_deprecation_log(config_filename, 'on_error'))
        config.setdefault('commands', []).append(
            {
                'after': 'error',
                'when': ['create', 'prune', 'compact', 'check'],
                'run': commands,
            },
        )

    # Normalize "before_everything" and "after_everything".
    for preposition in ('before', 'after'):
        option_name = f'{preposition}_everything'
        commands = config.pop(option_name, None)

        if commands:
            logs.append(make_command_hook_deprecation_log(config_filename, option_name))
            config.setdefault('commands', []).append(
                {
                    preposition: 'everything',
                    'when': ['create'],
                    'run': commands,
                },
            )

    return logs


def normalize(config_filename, config):  # noqa: PLR0912, PLR0915
    '''
    Given a configuration filename and a configuration dict of its loaded contents, apply particular
    hard-coded rules to normalize the configuration to adhere to the current schema. Return any log
    message warnings produced based on the normalization performed.

    Raise ValueError the configuration cannot be normalized.
    '''
    logs = normalize_sections(config_filename, config)
    logs += normalize_commands(config_filename, config)

    if config.get('borgmatic_source_directory'):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The borgmatic_source_directory option is deprecated and will be removed from a future release. Use user_runtime_directory and user_state_directory instead.',
                ),
            ),
        )

    # Upgrade exclude_if_present from a string to a list.
    exclude_if_present = config.get('exclude_if_present')
    if isinstance(exclude_if_present, str):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The exclude_if_present option now expects a list value. String values for this option are deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['exclude_if_present'] = [exclude_if_present]

    # Unconditionally set the bootstrap hook so that it's enabled by default and config files get
    # stored in each Borg archive.
    config.setdefault('bootstrap', {})

    # Move store_config_files from the global scope to the bootstrap hook.
    store_config_files = config.get('store_config_files')
    if store_config_files is not None:
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The store_config_files option has moved under the bootstrap hook. Specifying store_config_files at the global scope is deprecated and support will be removed from a future release.',
                ),
            ),
        )
        del config['store_config_files']
        config['bootstrap']['store_config_files'] = store_config_files

    # Upgrade various monitoring hooks from a string to a dict.
    healthchecks = config.get('healthchecks')
    if isinstance(healthchecks, str):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The Healthchecks hook now expects a key/value pair with "ping_url" as a key. String values for this option are deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['healthchecks'] = {'ping_url': healthchecks}

    cronitor = config.get('cronitor')
    if isinstance(cronitor, str):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The Cronitor hook now expects key/value pairs. String values for this option are deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['cronitor'] = {'ping_url': cronitor}

    pagerduty = config.get('pagerduty')
    if isinstance(pagerduty, str):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The PagerDuty hook now expects key/value pairs. String values for this option are deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['pagerduty'] = {'integration_key': pagerduty}

    # Upgrade consistency checks from a list of strings to a list of dicts.
    checks = config.get('checks')
    if isinstance(checks, list) and len(checks) and isinstance(checks[0], str):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The checks option now expects a list of key/value pairs. Lists of strings for this option are deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['checks'] = [{'name': check_type} for check_type in checks]

    # Rename various configuration options.
    numeric_owner = config.pop('numeric_owner', None)
    if numeric_owner is not None:
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The numeric_owner option has been renamed to numeric_ids. numeric_owner is deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['numeric_ids'] = numeric_owner

    bsd_flags = config.pop('bsd_flags', None)
    if bsd_flags is not None:
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The bsd_flags option has been renamed to flags. bsd_flags is deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['flags'] = bsd_flags

    remote_rate_limit = config.pop('remote_rate_limit', None)
    if remote_rate_limit is not None:
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The remote_rate_limit option has been renamed to upload_rate_limit. remote_rate_limit is deprecated and support will be removed from a future release.',
                ),
            ),
        )
        config['upload_rate_limit'] = remote_rate_limit

    # Upgrade remote repositories to ssh:// syntax, required in Borg 2.
    repositories = config.get('repositories')
    if repositories:
        if any(isinstance(repository, str) for repository in repositories):
            logs.append(
                logging.makeLogRecord(
                    dict(
                        levelno=logging.WARNING,
                        levelname='WARNING',
                        msg=f'{config_filename}: The repositories option now expects a list of key/value pairs. Lists of strings for this option are deprecated and support will be removed from a future release.',
                    ),
                ),
            )
            config['repositories'] = [
                {'path': repository} if isinstance(repository, str) else repository
                for repository in repositories
            ]
            repositories = config['repositories']

        config['repositories'] = []

        for repository_dict in repositories:
            repository_path = repository_dict.get('path')

            if repository_path is None:
                continue

            if '~' in repository_path:
                logs.append(
                    logging.makeLogRecord(
                        dict(
                            levelno=logging.WARNING,
                            levelname='WARNING',
                            msg=f'{config_filename}: Repository paths containing "~" are deprecated in borgmatic and support will be removed from a future release.',
                        ),
                    ),
                )

            if ':' in repository_path:
                if repository_path.startswith('file://'):
                    updated_repository_path = os.path.abspath(
                        repository_path.partition('file://')[-1],
                    )
                    config['repositories'].append(
                        dict(
                            repository_dict,
                            path=updated_repository_path,
                        ),
                    )
                elif repository_path.startswith(('ssh://', 'sftp://', 'rclone:', 's3:', 'b2:')):
                    config['repositories'].append(repository_dict)
                else:
                    rewritten_repository_path = f"ssh://{repository_path.replace(':~', '/~').replace(':/', '/').replace(':', '/./')}"
                    logs.append(
                        logging.makeLogRecord(
                            dict(
                                levelno=logging.WARNING,
                                levelname='WARNING',
                                msg=f'{config_filename}: Remote repository paths without ssh://, sftp://, rclone:, s3:, or b2:, syntax are deprecated and support will be removed from a future release. Interpreting "{repository_path}" as "{rewritten_repository_path}"',
                            ),
                        ),
                    )
                    config['repositories'].append(
                        dict(
                            repository_dict,
                            path=rewritten_repository_path,
                        ),
                    )
            else:
                config['repositories'].append(repository_dict)

    if config.get('prefix'):
        logs.append(
            logging.makeLogRecord(
                dict(
                    levelno=logging.WARNING,
                    levelname='WARNING',
                    msg=f'{config_filename}: The prefix option is deprecated and support will be removed from a future release. Use archive_name_format or match_archives instead.',
                ),
            ),
        )

    return logs



import logging

import requests

logger = logging.getLogger(__name__)


TIMEOUT_SECONDS = 10


def initialize_monitor(
    push_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Make a get request to the configured Uptime Kuma push_url. Use the given configuration filename
    in any log entries. If this is a dry run, then don't actually push anything.
    '''
    run_states = hook_config.get('states', ['start', 'finish', 'fail'])

    if state.name.lower() not in run_states:
        return

    dry_run_label = ' (dry run; not actually pushing)' if dry_run else ''
    status = 'down' if state.name.lower() == 'fail' else 'up'
    push_url = hook_config.get('push_url', 'https://example.uptime.kuma/api/push/abcd1234')
    query = f'status={status}&msg={state.name.lower()}'
    logger.info(f'Pushing Uptime Kuma push_url {push_url}?{query} {dry_run_label}')
    logger.debug(f'Full Uptime Kuma state URL {push_url}?{query}')

    if dry_run:
        return

    logging.getLogger('urllib3').setLevel(logging.ERROR)

    try:
        response = requests.get(
            f'{push_url}?{query}',
            verify=hook_config.get('verify_tls', True),
            timeout=TIMEOUT_SECONDS,
            headers={'User-Agent': 'borgmatic'},
        )
        if not response.ok:
            response.raise_for_status()
    except requests.exceptions.RequestException as error:
        logger.warning(f'Uptime Kuma error: {error}')


def destroy_monitor(push_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''



import collections
import contextlib
import enum
import json
import locale
import logging
import os
import select
import subprocess
import textwrap
import time

import borgmatic.logger

logger = logging.getLogger(__name__)


ERROR_OUTPUT_MAX_LINE_COUNT = 25
BORG_ERROR_EXIT_CODE_START = 2
BORG_ERROR_EXIT_CODE_END = 99

# See https://borgbackup.readthedocs.io/en/stable/internals/frontends.html#message-ids
BORG_WARNING_EXIT_CODES_TREATED_AS_ERRORS = {101, 102, 104, 105, 106}


class Exit_status(enum.Enum):
    STILL_RUNNING = 1
    SUCCESS = 2
    WARNING = 3
    ERROR = 4


def command_is_borg(command, borg_local_path):
    '''
    Given a command as a sequence and the Borg local path, return whether that command is a call to
    Borg.
    '''
    parsed_command = command.split(' ', 1) if isinstance(command, str) else command

    if not parsed_command:
        return False

    return bool(borg_local_path and parsed_command[0] == borg_local_path)


BORG_EXIT_CODE_TO_DESCRIPTION = {
    100: 'File changed while we backed it up',
    101: 'Include pattern never matched',
    102: 'General backup issue',
    103: 'File type or inode changed while we backed it up',
    104: 'Backup OS issue',
    105: 'Backup permission issue',
    106: 'Backup IO issue',
    107: 'Backup file not found',
}


def interpret_exit_code(command, exit_code, borg_local_path=None, borg_exit_codes=None):  # noqa: PLR0911
    '''
    Return an Exit_status value (e.g. SUCCESS, ERROR, or WARNING) based on interpreting the given
    exit code. If a Borg local path is given and matches the process' command, then interpret the
    exit code based on Borg's documented exit code semantics. And if Borg exit codes are given as a
    sequence of exit code configuration dicts, then take those configured preferences into account.
    '''
    if exit_code is None:
        return Exit_status.STILL_RUNNING

    if exit_code == 0:
        return Exit_status.SUCCESS

    if not command_is_borg(command, borg_local_path):
        return Exit_status.ERROR

    description = BORG_EXIT_CODE_TO_DESCRIPTION.get(exit_code)
    description_parenthetical = f' ({description})' if description else ''

    # First try looking for the exit code in the borg_exit_codes configuration.
    for entry in borg_exit_codes or ():
        if entry.get('code') == exit_code:
            treat_as = entry.get('treat_as')

            if treat_as == 'error':
                logger.error(
                    f'Treating exit code {exit_code}{description_parenthetical} as an error, as per configuration',
                )
                return Exit_status.ERROR

            if treat_as == 'warning':
                logger.warning(
                    f'Treating exit code {exit_code}{description_parenthetical} as a warning, as per configuration',
                )
                return Exit_status.WARNING

    # If the exit code doesn't have explicit configuration, then fall back to the default
    # behavior of treating Borg errors as errors and some Borg warnings as errors.
    if exit_code in BORG_WARNING_EXIT_CODES_TREATED_AS_ERRORS:
        logger.error(
            f'Treating exit code {exit_code}{description_parenthetical} as an error, as per borgmatic defaults',
        )

        return Exit_status.ERROR

    if exit_code < 0 or (
        exit_code >= BORG_ERROR_EXIT_CODE_START and exit_code <= BORG_ERROR_EXIT_CODE_END
    ):
        return Exit_status.ERROR

    logger.warning(
        f'Treating exit code {exit_code}{description_parenthetical} as a warning, as per borgmatic defaults',
    )
    return Exit_status.WARNING


def command_for_process(process):
    '''
    Given a process as an instance of subprocess.Popen, return the command string that was used to
    invoke it.
    '''
    return process.args if isinstance(process.args, str) else ' '.join(process.args)


def output_buffers_for_process(process, exclude_stdouts):
    '''
    Given a process as an instance of subprocess.Popen and a sequence of stdouts to exclude, return
    the process stdout and stderr as a tuple—but exclude the stdout if it's in the given stdouts to
    exclude.
    '''
    return tuple(
        buffer for buffer in (process.stdout, process.stderr) if buffer not in exclude_stdouts
    )


BORG_LOG_LEVEL_ELEVATION_THRESHOLD = 10


def borg_json_log_line_to_record(line, log_level):
    '''
    Given a single Borg "--log-json"-style log line and a log level, return the line converted to a
    logging.LogRecord instance. Return None if the line can't be parsed as JSON.

    If Borg provides a log level in its JSON, prefer logging at that level. But if Borg doesn't
    provide a log level—or the log level given to this function is just a little bit higher than
    Borg's—elevate to that level. This supports use cases like elevating Borg's INFO level logs to
    borgmatic's custom ANSWER level so that requested data shows up even at the default verbosity.
    '''
    with contextlib.suppress(json.JSONDecodeError, TypeError, KeyError, AttributeError):
        log_data = json.loads(line)
        log_type = log_data.get('type')

        if log_type == 'log_message':
            borg_log_level = logging._nameToLevel.get(log_data.get('levelname'))
            log_level_delta = 0 if log_level is None else log_level - borg_log_level

            if log_level_delta > 0 and log_level_delta < BORG_LOG_LEVEL_ELEVATION_THRESHOLD:
                return logging.makeLogRecord(
                    dict(
                        levelno=log_level,
                        created=log_data.get('time'),
                        msg=log_data.get('message'),
                        msgid=log_data.get('msgid'),
                        levelname=logging.getLevelName(log_level),
                        name=log_data.get('name'),
                    )
                )

            return logging.makeLogRecord(
                dict(
                    levelno=borg_log_level,
                    created=log_data.get('time'),
                    msg=log_data.get('message'),
                    msgid=log_data.get('msgid'),
                    levelname=log_data.get('levelname'),
                    name=log_data.get('name'),
                )
            )

        if log_type == 'file_status':
            return logging.makeLogRecord(
                dict(
                    levelno=log_level,
                    created=time.time(),
                    msg=f'{log_data.get("status")} {log_data.get("path")}',
                    msgid=log_data.get('msgid'),
                    levelname=logging.getLevelName(log_level),
                    name='borg.file_status',
                )
            )

    return None


def log_line_to_record(line, log_level):
    '''
    Given a log data dict for a single Borg log entry and a log level, return it converted to a
    logging.LogRecord instance.
    '''
    return logging.makeLogRecord(
        dict(
            msg=line,
            levelno=log_level,
            levelname=logging.getLevelName(log_level),
        )
    )


def parse_log_line(line, log_level, elevate_stderr, borg_local_path, command):
    '''
    Given a raw output line from an external program, whether this line came from stderr and should
    be elevated to error/warning, the Borg local path, and the command as a sequence, return a
    logging.LogRecord instance containing its parsed data.

    If the command being run is Borg, and the log line is JSON-formatted log data, then grab the log
    level from it and log the parsed JSON to be consumed later by a Python logging.Formatter.

    But for non-Borg commands, elevate stderr-sourced logs to ERROR. The one exception is if the log
    came from stderr and the string "warning:" appears at the start of the log line. In that case,
    just elevate the log level to a WARN.
    '''
    if command_is_borg(command, borg_local_path):
        log_record = borg_json_log_line_to_record(line, log_level)

        if log_record:
            return log_record

    if elevate_stderr:
        return log_line_to_record(
            line, logging.WARNING if line.lower().startswith('warning:') else logging.ERROR
        )

    return log_line_to_record(line, log_level)


def handle_log_record(log_record, last_lines=None):
    '''
    Given a log record to be logged and a rolling list of last lines, append the record's message to
    the last lines (if given). Then (if the log level is not None), log the record.

    Return the log record.
    '''
    log_message = log_record.getMessage()

    if last_lines is not None:
        last_lines.append(log_message)

        if len(last_lines) > ERROR_OUTPUT_MAX_LINE_COUNT:
            last_lines.pop(0)

    if log_record.levelno is not None:
        logger.handle(log_record)

    return log_record


READ_CHUNK_SIZE = 4096


def read_lines(buffer, process, line_separator='\n'):
    '''
    Given a Python buffer (like stdout) ready for reading, its process, and a line separator,
    repeatedly yield a tuple of (decoded) lines from the buffer until the process has exited.

    It is assumed that this function's generator is used in conjunction with an external select()
    call to know when to read more lines. Otherwise, the generator will busywait if it's called in a
    tight loop.
    '''
    data = b''
    encoded_separator = line_separator.encode()
    separator_size = len(encoded_separator)
    encoding = locale.getpreferredencoding()

    while True:
        chunk = os.read(buffer.fileno(), READ_CHUNK_SIZE)

        if not chunk:  # EOF
            # The process is still running, so we keep running too.
            if process.poll() is None:  # pragma: no cover
                continue

            break

        data += chunk
        lines = []

        # Split the data into lines, holding back anything leftover that might
        # be a partial line.
        while True:
            separator_position = data.find(encoded_separator)

            if separator_position == -1:
                break

            lines.append(data[:separator_position].decode(encoding))
            data = data[separator_position + separator_size :]

        yield tuple(lines)

    # Yield any leftover data from the end of the buffer.
    if data:
        yield (data.decode(encoding).rstrip(),)


Buffer_reader = collections.namedtuple(
    'Buffer_reader',
    ('lines', 'process'),
)


Process_metadata = collections.namedtuple(
    'Process_metadata',
    ('last_lines', 'capture'),
)


def log_buffer_lines(
    buffer_readers, process_metadatas, output_log_level, borg_local_path, capture_stderr=False
):
    '''
    Given a dict from buffer object to Buffer_reader, a dict from subprocess.Popen() instance to
    Process_metadata instance, a requested output log level for stdout, Borg's local path, and
    whether to capture stderr, read and log any ready output lines from the buffers. Additionally,
    for any log records with a log level the same as the output log level, yield those log messages
    for capture.

    This function just does one "turn of the crank" of logging buffer output. It is intended to be
    called repeatedly to continue to process buffers.
    '''
    if not buffer_readers:
        return

    (ready_buffers, _, _) = select.select(buffer_readers.keys(), [], [])

    for ready_buffer in ready_buffers:
        reader = buffer_readers[ready_buffer]

        # The "ready" process has exited, but it might be a pipe destination with other
        # processes (pipe sources) waiting to be read from. So as a measure to prevent
        # hangs, vent all processes when one exits.
        if reader.process and reader.process.poll() is not None:
            for other_process in process_metadatas:
                if (
                    other_process.poll() is None
                    and other_process.stdout
                    and other_process.stdout not in buffer_readers
                ):
                    # Add the process's output to buffer_readers to ensure it'll get read.
                    buffer_readers[other_process.stdout] = Buffer_reader(
                        read_lines(other_process.stdout, other_process), other_process
                    )

        try:
            lines = next(reader.lines)
        except StopIteration:
            continue

        for line in lines:
            if not line or not reader.process:
                continue

            # Keep the last few lines of output in case the process errors and we need the
            # output for the exception below.
            log_record = handle_log_record(
                parse_log_line(
                    line=line,
                    log_level=output_log_level,
                    elevate_stderr=(ready_buffer == reader.process.stderr and not capture_stderr),
                    borg_local_path=borg_local_path,
                    command=reader.process.args,
                ),
                last_lines=process_metadatas[reader.process].last_lines,
            )

            if (
                log_record.levelno is None or log_record.levelno == output_log_level
            ) and process_metadatas[reader.process].capture:
                yield log_record.getMessage()


def raise_for_process_errors(
    buffer_readers,
    process_metadatas,
    output_log_level,
    borg_local_path,
    borg_exit_codes,
    capture_stderr=False,
):
    '''
    Given a dict from buffer object to Buffer_reader, a dict from subprocess.Popen() instance to
    Process_metadata instance, a requested output log level for stdout, Borg's local path, a
    sequence of exit code configuration dicts, and whether to capture stderr, check the given
    processes for error or warning exit codes. If found, vent or kill any running processes and
    drain any remaining buffer lines. In the case of an error exit code, raise.  In the case of
    warning, return Exit_status.WARNING. Otherwise, return None.
    '''
    result_status = None

    for process in process_metadatas:
        exit_code = process.poll() if buffer_readers else process.wait()

        if exit_code is None:
            continue

        exit_status = interpret_exit_code(process.args, exit_code, borg_local_path, borg_exit_codes)

        if exit_status not in {Exit_status.ERROR, Exit_status.WARNING}:
            continue

        # Something has gone wrong. So vent each process' output buffer to prevent it from
        # hanging. And then kill the process.
        for other_process in process_metadatas:
            if other_process.poll() is None:
                other_process.stdout.read(0)
                other_process.kill()

        if exit_status == Exit_status.WARNING:
            result_status = Exit_status.WARNING
            continue

        # Drain and log remaining buffer lines, so no output gets lost. But swallow captured output,
        # since we're raising below instead of yielding captured lines.
        tuple(
            log_remaining_buffer_lines(
                buffer_readers, process_metadatas, output_log_level, borg_local_path, capture_stderr
            )
        )
        last_lines = process_metadatas[process].last_lines

        # If an error occurs, include its output in the raised exception so that we don't
        # inadvertently hide error output.
        if len(last_lines) >= ERROR_OUTPUT_MAX_LINE_COUNT:
            last_lines.insert(0, '...')

        raise subprocess.CalledProcessError(
            exit_code,
            command_for_process(process),
            '\n'.join(last_lines),
        )

    return result_status


def log_remaining_buffer_lines(
    buffer_readers, process_metadatas, output_log_level, borg_local_path, capture_stderr=False
):
    '''
    Given a dict from buffer object to Buffer_reader, a dict from subprocess.Popen() instance to
    Process_metadata instance, a requested output log level for stdout, Borg's local path, and
    whether to capture stderr, drain and log any remaining output lines from the buffers until
    they're empty. Additionally, for any log records with a log level the same as the output log
    level, yield those log messages for capture.

    The main difference between this function and log_buffer_lines() is that this one doesn't just
    do one "turn of the crank" of logging buffer output; it completely drains any remaining output.
    '''
    for output_buffer, reader in buffer_readers.items():
        if not reader.process:
            continue

        for lines in reader.lines:
            for line in lines:
                log_record = handle_log_record(
                    parse_log_line(
                        line=line.rstrip(),
                        log_level=output_log_level,
                        elevate_stderr=(
                            output_buffer == reader.process.stderr and not capture_stderr
                        ),
                        borg_local_path=borg_local_path,
                        command=reader.process.args,
                    ),
                    last_lines=process_metadatas[reader.process].last_lines,
                )

                if (
                    log_record.levelno is None or log_record.levelno == output_log_level
                ) and process_metadatas[reader.process].capture:
                    yield log_record.getMessage()


def log_outputs(
    processes,
    exclude_stdouts,
    output_log_level,
    borg_local_path,
    borg_exit_codes,
    capture_stderr=False,
):
    '''
    Given a sequence of subprocess.Popen() instances for multiple processes, log the outputs (stderr
    and stdout). Use the requested output log level for stdout, but always log stderr to the ERROR
    log level. Additionally, raise a CalledProcessError if a process exits with an error (or a
    warning for exit code 1, if that process does not match the Borg local path).

    If the output log level is None, then instead of logging, capture the output for the last
    process given and yield it one line at a time. This includes stderr if capture stderr is set.
    But if the output log level is not None, don't yield anything.

    This yielding means that this function is a generator, and must be consumed in order to execute.

    Use the given Borg local path and exit code configuration to decide what's an error and what's a
    warning. If any stdouts are given to exclude, then for any matching processes, ignore those
    buffers. Also note that stdout for a process can be None if output is intentionally not
    captured, in which case it won't be logged.
    '''
    # Map from output buffer to Process_metadata instance. By convention, the last process is the
    # process to capture.
    process_metadatas = {
        process: Process_metadata(last_lines=[], capture=bool(process == processes[-1]))
        for process in processes
    }

    # Map from buffer to Buffer_reader instance.
    buffer_readers = {
        buffer: Buffer_reader(read_lines(buffer, process), process)
        for process in processes
        if process.stdout or process.stderr
        for buffer in output_buffers_for_process(process, exclude_stdouts)
    }

    # Log output lines for each process until they all exit or one errors.
    while True:
        yield from log_buffer_lines(
            buffer_readers, process_metadatas, output_log_level, borg_local_path, capture_stderr
        )

        if (
            raise_for_process_errors(
                buffer_readers,
                process_metadatas,
                output_log_level,
                borg_local_path,
                borg_exit_codes,
                capture_stderr,
            )
            == Exit_status.WARNING
        ):
            break

        if all(process.poll() is not None for process in processes):
            break

    # Now that all processes have exited, drain and consume any last output.
    yield from log_remaining_buffer_lines(
        buffer_readers, process_metadatas, output_log_level, borg_local_path, capture_stderr
    )


SECRET_COMMAND_FLAG_NAMES = {'--password'}


def mask_command_secrets(full_command):
    '''
    Given a command as a sequence, mask secret values for flags like "--password" in preparation for
    logging.
    '''
    masked_command = []
    previous_piece = None

    for piece in full_command:
        masked_command.append('***' if previous_piece in SECRET_COMMAND_FLAG_NAMES else piece)
        previous_piece = piece

    return tuple(masked_command)


MAX_LOGGED_COMMAND_LENGTH = 1000


PREFIXES_OF_ENVIRONMENT_VARIABLES_TO_LOG = ('BORG_', 'PG', 'MARIADB_', 'MYSQL_')


def log_command(full_command, input_file=None, output_file=None, environment=None):
    '''
    Log the given command (a sequence of command/argument strings), along with its input/output file
    paths and extra environment variables (with omitted values in case they contain passwords).
    '''
    logger.debug(
        textwrap.shorten(
            ' '.join(
                tuple(
                    f'{key}=***'
                    for key in (environment or {})
                    if any(
                        key.startswith(prefix)
                        for prefix in PREFIXES_OF_ENVIRONMENT_VARIABLES_TO_LOG
                    )
                )
                + mask_command_secrets(full_command),
            ),
            width=MAX_LOGGED_COMMAND_LENGTH,
            placeholder=' ...',
        )
        + (f" < {getattr(input_file, 'name', input_file)}" if input_file else '')
        + (f" > {getattr(output_file, 'name', output_file)}" if output_file else ''),
    )


# A sentinel passed as an output file to execute_command() to indicate that the command's output
# should be allowed to flow through to stdout without being captured for logging. Useful for
# commands with interactive prompts or those that mess directly with the console.
DO_NOT_CAPTURE = object()


def execute_command(
    full_command,
    output_log_level=logging.INFO,
    output_file=None,
    input_file=None,
    shell=False,
    environment=None,
    working_directory=None,
    borg_local_path=None,
    borg_exit_codes=None,
    run_to_completion=True,
    close_fds=False,  # Necessary for passing credentials via anonymous pipe.
):
    '''
    Execute the given command (a sequence of command/argument strings) and log its stdout output at
    the given log level. If an open output file object is given, then write stdout to the file and
    only log stderr. If an open input file object is given, then read stdin from the file. If shell
    is True, execute the command within a shell. If an environment variables dict is given, then
    pass it into the command. If a working directory is given, use that as the present working
    directory when running the command. If a Borg local path is given, and the command matches it
    (regardless of arguments), treat exit code 1 as a warning instead of an error. But if Borg exit
    codes are given as a sequence of exit code configuration dicts, then use that configuration to
    decide what's an error and what's a warning. If run to completion is False, then return the
    process for the command without executing it to completion.

    Raise subprocesses.CalledProcessError if an error occurs while running the command.
    '''
    log_command(full_command, input_file, output_file, environment)
    do_not_capture = bool(output_file is DO_NOT_CAPTURE)
    command = ' '.join(full_command) if shell else full_command

    process = subprocess.Popen(  # noqa: S603
        command,
        stdin=input_file,
        stdout=None if do_not_capture else (output_file or subprocess.PIPE),
        stderr=None if do_not_capture else subprocess.PIPE,
        shell=shell,
        env=environment,
        cwd=working_directory,
        close_fds=close_fds,
    )
    if not run_to_completion:
        return process

    with borgmatic.logger.Log_prefix(None):  # Log command output without any prefix.
        tuple(
            log_outputs(
                (process,),
                (input_file, output_file),
                output_log_level,
                borg_local_path,
                borg_exit_codes,
            )
        )

    return None


def execute_command_and_capture_output(
    full_command,
    output_log_level=None,
    input_file=None,
    capture_stderr=False,
    shell=False,
    environment=None,
    working_directory=None,
    borg_local_path=None,
    borg_exit_codes=None,
    close_fds=False,  # Necessary for passing credentials via anonymous pipe.
):
    '''
    Execute the given command (a sequence of command/argument strings), capturing and returning its
    output (stdout) as a generator that yields one line at a time. The generator must be consumed in
    order for the called command to execute.

    If an output log level is given, then instead of suppressing log output, also output the
    captured lines at the given log level.  If an input file descriptor is given, then pipe it to
    the command's stdin. If capture stderr is True, then capture stderr in addition to stdout. If
    shell is True, execute the command within a shell. If an environment variables dict is given,
    then pass it into the command. If a working directory is given, use that as the present working
    directory when running the command. If a Borg local path is given, and the command matches it
    (regardless of arguments), treat exit code 1 as a warning instead of an error. But if Borg exit
    codes are given as a sequence of exit code configuration dicts, then use that configuration to
    decide what's an error and what's a warning.

    Raise subprocesses.CalledProcessError if an error occurs while running the command.
    '''
    log_command(full_command, input_file, environment=environment)
    command = ' '.join(full_command) if shell else full_command

    try:
        process = subprocess.Popen(  # noqa: S603
            command,
            stdin=input_file,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
            if capture_stderr or command_is_borg(command, borg_local_path)
            else None,
            shell=shell,
            env=environment,
            cwd=working_directory,
            close_fds=close_fds,
        )
    except subprocess.CalledProcessError as error:
        if (
            interpret_exit_code(command, error.returncode, borg_local_path, borg_exit_codes)
            == Exit_status.ERROR
        ):
            raise

        if error.output is not None:
            yield from iter(error.output.decode(locale.getpreferredencoding()).splitlines())

        return

    with borgmatic.logger.Log_prefix(None):  # Log command output without any prefix.
        yield from log_outputs(
            (process,),
            (input_file,),
            output_log_level,
            borg_local_path,
            borg_exit_codes,
            capture_stderr=capture_stderr,
        )


def execute_command_with_processes(
    full_command,
    processes,
    output_log_level=logging.INFO,
    output_file=None,
    input_file=None,
    shell=False,
    environment=None,
    working_directory=None,
    borg_local_path=None,
    borg_exit_codes=None,
    close_fds=False,  # Necessary for passing credentials via anonymous pipe.
):
    '''
    Execute the given command (a sequence of command/argument strings) and log its stdout output at
    the given log level. Simultaneously, continue to poll one or more active processes so that they
    run as well. This is useful, for instance, for processes that are streaming output to a named
    pipe that the given command is consuming from.

    If an open output file object is given, then write stdout to the file and only log stderr. But
    if output log level is None, instead suppress logging and return the captured output for (only)
    the given command as a generator that yields one line at a time. The generator must be consumed
    in order for the called command to execute—regardless of the output log level.

    If an open input file object is given, then read stdin from the file. If shell is True, execute
    the command within a shell. If an environment variables dict is given, then pass it into the
    command. If a working directory is given, use that as the present working directory when running
    the command. If a Borg local path is given, then for any matching command or process (regardless
    of arguments), treat exit code 1 as a warning instead of an error. But if Borg exit codes are
    given as a sequence of exit code configuration dicts, then use that configuration to decide
    what's an error and what's a warning.

    Raise subprocesses.CalledProcessError if an error occurs while running the command or in the
    upstream process.
    '''
    log_command(full_command, input_file, output_file, environment)
    do_not_capture = bool(output_file is DO_NOT_CAPTURE)
    command = ' '.join(full_command) if shell else full_command

    try:
        command_process = subprocess.Popen(  # noqa: S603
            command,
            stdin=input_file,
            stdout=None if do_not_capture else (output_file or subprocess.PIPE),
            stderr=None if do_not_capture else subprocess.PIPE,
            shell=shell,
            env=environment,
            cwd=working_directory,
            close_fds=close_fds,
        )
    except (subprocess.CalledProcessError, OSError):
        # Something has gone wrong. So vent each process' output buffer to prevent it from hanging.
        # And then kill the process.
        for process in processes:
            if process.poll() is None:
                process.stdout.read(0)
                process.kill()

        raise

    with borgmatic.logger.Log_prefix(None):  # Log command output without any prefix.
        yield from log_outputs(
            (*processes, command_process),
            (input_file, output_file),
            output_log_level,
            borg_local_path,
            borg_exit_codes,
        )



import logging
import re
import urllib

import requests

logger = logging.getLogger(__name__)


TIMEOUT_SECONDS = 10


def initialize_monitor(
    ping_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


DATA_SOURCE_NAME_URL_PATTERN = re.compile(
    r'^(?P<protocol>.+)://(?P<username>.+)@(?P<hostname>.+)/(?P<project_id>.+)$',
)


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Construct and ping a Sentry cron URL, based on the configured DSN URL and monitor slug. Use the
    given configuration filename in any log entries. If this is a dry run, then don't actually ping
    anything.
    '''
    run_states = hook_config.get('states', ['start', 'finish', 'fail'])

    if state.name.lower() not in run_states:
        return

    dry_run_label = ' (dry run; not actually pinging)' if dry_run else ''

    data_source_name_url = hook_config.get('data_source_name_url')
    environment = hook_config.get('environment')
    monitor_slug = hook_config.get('monitor_slug')
    match = DATA_SOURCE_NAME_URL_PATTERN.match(data_source_name_url)

    if not match:
        logger.warning(f'Invalid Sentry data source name URL: {data_source_name_url}')
        return

    cron_url = f'{match.group("protocol")}://{match.group("hostname")}/api/{match.group("project_id")}/cron/{monitor_slug}/{match.group("username")}/'

    logger.info(f'Pinging Sentry {state.name.lower()}{dry_run_label}')
    logger.debug(f'Using Sentry cron URL {cron_url}')

    status = {
        'start': 'in_progress',
        'finish': 'ok',
        'fail': 'error',
    }.get(state.name.lower())

    if not status:
        logger.warning('Invalid Sentry state')
        return

    if dry_run:
        return

    environment_query = f'&environment={urllib.parse.quote(environment)}' if environment else ''

    logging.getLogger('urllib3').setLevel(logging.ERROR)
    try:
        response = requests.post(
            f'{cron_url}?status={status}{environment_query}',
            timeout=TIMEOUT_SECONDS,
            headers={'User-Agent': 'borgmatic'},
        )
        if not response.ok:
            response.raise_for_status()
    except requests.exceptions.RequestException as error:
        logger.warning(f'Sentry error: {error}')


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''



def repository_enabled_for_checks(repository, config):
    '''
    Given a repository name and a configuration dict, return whether the
    repository is enabled to have consistency checks run.
    '''
    if not config.get('check_repositories'):
        return True

    return repository in config['check_repositories']



import os
import re

VARIABLE_PATTERN = re.compile(
    r'(?P<escape>\\)?(?P<variable>\$\{(?P<name>[A-Za-z0-9_]+)((:?-)(?P<default>[^}]+))?\})',
)


def resolve_string(matcher):
    '''
    Given a matcher containing a name and an optional default value, get the value from environment.

    Raise ValueError if the variable is not defined in environment and no default value is provided.
    '''
    if matcher.group('escape') is not None:
        # In the case of an escaped environment variable, unescape it.
        return matcher.group('variable')

    # Resolve the environment variable.
    name, default = matcher.group('name'), matcher.group('default')
    out = os.getenv(name, default=default)

    if out is None:
        raise ValueError(f'Cannot find variable {name} in environment')

    return out


def resolve_env_variables(item):
    '''
    Resolves variables like or ${FOO} from given configuration with values from process environment.

    Supported formats:

     * ${FOO} will return FOO env variable
     * ${FOO-bar} or ${FOO:-bar} will return FOO env variable if it exists, else "bar"

    Raise if any variable is missing in environment and no default value is provided.
    '''
    if isinstance(item, str):
        return VARIABLE_PATTERN.sub(resolve_string, item)

    if isinstance(item, list):
        for index, subitem in enumerate(item):
            item[index] = resolve_env_variables(subitem)

    if isinstance(item, dict):
        for key, value in item.items():
            item[key] = resolve_env_variables(value)

    return item



import logging

import borgmatic.config.generate
import borgmatic.config.validate
import borgmatic.logger

logger = logging.getLogger(__name__)


def run_generate(generate_arguments, global_arguments):
    '''
    Given the generate arguments and the global arguments, each as an argparse.Namespace instance,
    run the "generate" action.

    Raise FileExistsError if a file already exists at the destination path and the generate
    arguments do not have overwrite set.
    '''
    borgmatic.logger.add_custom_log_levels()
    dry_run_label = ' (dry run; not actually writing anything)' if global_arguments.dry_run else ''

    logger.answer(
        f'Generating configuration files within: {generate_arguments.destination_path}{dry_run_label}'
        if generate_arguments.split
        else f'Generating a configuration file at: {generate_arguments.destination_path}{dry_run_label}'
    )

    borgmatic.config.generate.generate_sample_configuration(
        global_arguments.dry_run,
        generate_arguments.source_filename,
        generate_arguments.destination_path,
        borgmatic.config.validate.schema_filename(),
        overwrite=generate_arguments.overwrite,
        split=generate_arguments.split,
    )

    if generate_arguments.source_filename:
        logger.answer(
            f'''
Merged the contents of the source configuration file: {generate_arguments.source_filename}'''
        )

        if not generate_arguments.split:
            logger.answer(
                f'''To review the changes made, run:

    diff --unified {generate_arguments.source_filename} {generate_arguments.destination_path}''',
            )

    logger.answer(
        '''
This includes all available configuration options with example values, the few
required options as indicated. Please edit the file to suit your needs.

If you ever need help: https://torsion.org/borgmatic/#issues''',
    )
